from compas_tno.problems import initialize_loadpath


def initialize_loadpath_proxy(formdata, problem=None):
    """ Initialise loadpath using the proxy.

    Parameters
    ----------
    formdata : dict
        Data of the form diagram
    problem : :class:`~compas_tno.problems.Problem`, optional
        Class with matrices of the problem, by default None

    Returns
    -------
    formdata : dict
        Data of the optimised form diagram
    output : dict
        Dictionary with the values of the optimisation
    """

    from compas_tno.diagrams import FormDiagram

    form = FormDiagram.from_data(formdata)
    output = initialize_loadpath(form, problem=problem)

    # TODO: add an option here to initialise the loadpath with SLSQP/IPOPT if matlab/cvx can not be reached.

    return form.to_data(), output


def run_NLP_proxy(shapedata, formdata, optimiserdata):
    """Run nonlinear multiobjective optimisations using the proxy

    Parameters
    ----------
    shapedata : dict
        Data of the Shape object
    formdata : dict
        Data of the Form object
    optimiserdata : dict
        Data of the Optimiser object

    Returns
    -------
    shapedata : dict
        Data of the Shape object
    formdata : dict
        Data of the optimised form diagram.
    optimiserdata : dict
        Data of the Optimiser object.
    """

    from compas_tno.diagrams import FormDiagram
    from compas_tno.analysis import Analysis
    from compas_tno.optimisers import Optimiser
    from compas_tno.shapes import Shape

    shape = Shape.from_data(shapedata)
    form = FormDiagram.from_data(formdata)
    optimiser = Optimiser.from_data(optimiserdata)

    analysis = Analysis.from_elements(shape, form, optimiser)
    analysis.set_up_optimiser()
    analysis.run()

    return shape.to_data(), form.to_data(), optimiser.to_data()




from numpy import arange
from numpy import array
from numpy import append
from numpy import linspace

from compas.geometry import intersection_line_line_xy

import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter
from matplotlib import cm

import compas_tno
import json
import csv
import os


__all__ = [
    'diagram_of_thrust',
    'diagram_of_multiple_thrust',
    'diagram_of_thrust_load_mult',
    'surface_GSF_load_mult',
    'save_csv_row',
    'open_csv_row',
    'interpolate_min_thk',
    'filter_min_thk',
    'lookup_folder',
    'save_pointcloud'
]


def diagram_of_thrust(thicknesses, solutions, limit_state=True, fill=False, xy_limits=None, GSF_ticks=False, save=False):
    """ Plot a diagram of Thrusts based on the collected data from (n) points.

    Parameters
    ----------
    thicknesses : list of lists [[n],[m]]
        Points with discretised solutions for minimum/maximum thrust.
    solutions : list of lists [[n],[m]]
        Adimensional thrust over weight for minimum/maximum thrust.
    limit_state : bool, optional
        If yes, it interpolates the limit_state based on the two last solutions.
        The default value is ``True``.
    fill : bool, optional
        If yes, it fills bewteen the curves of min and max.
        The default value is ``False``.
    xy_limits : list of lists, optional
        Freezes the axis of the graphs.
        The default value is ``None`` in which case these limits are calculated automarically
    GSF_ticks : bool, optional
        Add the ticks regarding the GSF to the top axis.
        The default value is ``False``.
    save : str, optional
        Path to save the graph.
        The default value is ``False``.

    Returns
    -------
    obj
        Plotter object.

    """
    thicknesses_min, thicknesses_max = thicknesses
    min_sol, max_sol = solutions
    xmin = thicknesses_min
    xmax = thicknesses_max
    fmin = 100.0 * array(min_sol)
    fmax = 100.0 * array(max_sol)  # add minus signe
    # print('\n', xmin, xmax, fmin, fmax)
    n = len(xmin)
    m = len(xmax)
    if m > n:
        dimension = xmax
    else:
        dimension = xmin

    if xmin[-1] == xmax[-1]:
        x_ = xmin[-1]
        y_ = fmin[-1]
        limit_state = True
        dim = dimension
        xmax_ = xmax
        xmin_ = xmin
        fmin_ = fmin
        fmax_ = fmax
    else:
        if n >= 2 and m >= 2:
            x_, y_, _ = intersection_line_line_xy([[xmax[m-1], fmax[m-1]], [xmax[m-2], fmax[m-2]]], [[xmin[n-1], fmin[n-1]], [xmin[n-2], fmin[n-2]]])
        dim = append(dimension, x_)
        xmax_ = append(xmax, x_)
        xmin_ = append(xmin, x_)
        fmin_ = append(fmin, y_)
        fmax_ = append(fmax, y_)

    # print('\n', xmin_, xmax_, fmin_, fmax_)

    # size_axis_label = 14
    # size_axis_data = 12
    # size_legend = 14
    size_axis_label = 12
    size_axis_data = 12
    size_legend = 12

    interval_x = abs(dimension[0] - dimension[1])
    interval_y = 10
    if xy_limits:
        [[max_x, min_x], [max_y, min_y]] = xy_limits
    else:
        max_x = max(dim)
        min_x = min(dim) - interval_x
        max_y = interval_y - max(fmax_) % 10 + max(fmax_)
        min_y = min(fmin_) - min(fmin_) % 10

    if not GSF_ticks:
        last_scale = round(max_x/min_x, 2)
        middle_scale = round((last_scale - 1.0)/2 + 1, 2)
        quad_scale = round((last_scale - 1.0)*3/4 + 1, 2)
        ticks_GSF = [middle_scale, quad_scale, last_scale]
    else:
        ticks_GSF = GSF_ticks

    ticks_x = []
    for tck in ticks_GSF:
        ticks_x.append(max_x/tck)

    fig = plt.figure(figsize=[12, 4])

    ax = fig.add_subplot(1, 1, 1)
    ax.plot(xmin, fmin, 'o', ls='-', markersize=5, color='blue', label='minimum thrust')
    ax.plot(xmax, fmax, 'o', ls='-', markersize=5, color='red', label='maximum thrust')

    if limit_state:
        extrapolation_max = [fmax[m-1], y_]
        extrapolation_min = [fmin[n-1], y_]
        extrapolation_x_max = [xmax[m-1], x_]
        extrapolation_x_min = [xmin[n-1], x_]
        ax.plot(x_, y_, 'o', ls=' ', markersize=7, color='black', label='limit state')
        ax.plot(extrapolation_x_max, extrapolation_max, '', ls='--', color='red')
        ax.plot(extrapolation_x_min, extrapolation_min, '', ls='--', color='blue')
        ax.annotate(str(round(max_x/x_, 2)), (x_, y_), textcoords="offset points", xytext=(0, 10), ha='center')

    if fill:
        ax.fill(append(xmin_, xmax_[::-1]), append(fmin_, fmax_[::-1]), color="grey", alpha=0.2)

    ax1 = plt.axes()
    ax2 = ax1.twiny()

    ax2.set_xticks([0] + [100*(max_x-tck_x)/(max_x-min_x) for tck_x in ticks_x])
    ax2.set_xticklabels(['1.0'] + [str(round(tck_GSF, 2)) for tck_GSF in ticks_GSF], size=size_axis_data)

    ax1.set_xlabel('thickness/radius', size=size_axis_label, weight='bold', labelpad=8)
    ax2.set_xlabel('GSF', size=size_axis_label, weight='bold', labelpad=8)
    ax1.set_ylabel('thrust/weight [%]', size=size_axis_label, weight='bold', labelpad=8)
    # ax1.xaxis.set_major_formatter(FormatStrFormatter('%.2f'))  # Check if this is necessary
    ax1.yaxis.set_major_formatter(FormatStrFormatter('%.0f'))
    ax1.set_xlim(max_x, min_x)
    ax.set_ylim(min_y, max_y)
    # ax1.set_xticks(arange(max_x, min_x - interval_x, -interval_x))
    ax1.set_yticks(arange(min_y, max_y + interval_y, interval_y))
    ax1.tick_params(axis='both', which='major', labelsize=size_axis_data)

    # box = ax.get_position()
    # ax.set_position([box.x0*0.6, box.y0*1.5, box.width * 0.90, box.height*0.90])
    ax.grid(color='silver', linestyle='-', linewidth=0.5)
    # ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), fontsize=size_legend)  # , bbox_to_anchor(0.1, 0.1), ncol=1)
    ax.legend(fontsize=size_legend)  # , bbox_to_anchor(0.1, 0.1), ncol=1)

    if save:
        plt.savefig(save)

    return plt


def diagram_of_multiple_thrust(thicknesses, solutions, legends, simplified=True, limit_state=True, colors=None, xy_limits=None, GSF_ticks=False,
                               fill=False, show_legend=True, save=None, markers=None):
    """ Plot a diagram of Thrusts based on the collected data on (m) problems each with (n) points.

    Parameters
    ----------
    thicknesses : list of lists [[n],[m]]
        Points with discretised solutions for minimum/maximum thrust.
    solutions : list of lists [[n],[m]]
        Adimensional thrust over weight for minimum/maximum thrust.
    legends : list
        A list with the name of each data entry
    simplified : bool, optional
        If yes, the maximum and minimum thrust series have the same color.
        The default value is ``True``.
    limit_state : bool, optional
        If yes, it interpolates the limit_state based on the two last solutions.
        The default value is ``True``.
    colors : list, optional
        If added, this define the HEX color for each of the diagrams to plot.
        The default value is ``True``.
    xy_limits : list of lists, optional
        Freezes the axis of the graphs.
        The default value is ``None`` in which case these limits are calculated automarically
    GSF_ticks : bool, optional
        Add the ticks regarding the GSF to the top axis.
        The default value is ``False``.
    fill : bool, optional
        If yes, it fills bewteen the curves of min and max.
        The default value is ``False``.
    show_legend : bool, optional
        Defines if the legend is plotted in the graph.
        The default value is ``True``.
    markers : list, optional
        If added, this define the markers for each diagram.
        The default value is ``True``.
    save : str, optional
        Path to save the graph.
        The default value is ``False``.

    Returns
    -------
    obj
        Plotter object.

    """

    kmax = len(legends)

    if markers is None:
        markers = ['o', '^', 's', 'D', 'x', '1', '2', '3', 'v', 'p', '4', '8']

    if colors is None:
        colormap = plt.cm.get_cmap('coolwarm')  # gist_ncar nipy_spectral, Set1, Paired coolwarm
        colors = [colormap(i) for i in linspace(0, 1.0, kmax)]

    size_axis_label = 14
    size_axis_data = 12
    size_legend = 14

    fig = plt.figure(figsize=[15, 5])
    ax = fig.add_subplot(1, 1, 1)

    interval_x = 0.05  # abs(flatten_list(min_thks)[0] - flatten_list(min_thks)[1])
    interval_y = 20

    if xy_limits is None:
        min_sols = solutions[0][0]  # this has to find the maximum from all the input series
        max_sols = solutions[0][1]
        min_thks = thicknesses[0][1]
        def flatten_list(li): return [item for sublist in li for item in sublist]
        extreme_max = -100*min(flatten_list(max_sols))
        extreme_min = 100*min(flatten_list(min_sols))
        max_x = max(flatten_list(min_thks))
        min_x = min(flatten_list(min_thks)) - interval_x
        max_y = interval_y - extreme_max % 10 + extreme_max
        min_y = extreme_min - extreme_min % 10
    else:
        [[max_x, min_x], [max_y, min_y]] = xy_limits

    if not GSF_ticks:
        last_scale = round(max_x/min_x, 2)
        middle_scale = round((last_scale - 1.0)/2 + 1, 2)
        quad_scale = round((last_scale - 1.0)*3/4 + 1, 2)
        ticks_GSF = [middle_scale, quad_scale, last_scale]
    else:
        ticks_GSF = GSF_ticks

    ticks_x = []
    for tck in ticks_GSF:
        ticks_x.append(max_x/tck)

    # print(thicknesses)
    # print(solutions)
    # print(kmax)
    for i in range(kmax):
        thks = thicknesses[i]
        sols = solutions[i]

        xmin = array(thks[0])
        xmax = array(thks[1])

        fmin = 100.0*array(sols[0])
        fmax = -100.0*array(sols[1])
        n = len(xmin)
        m = len(xmax)

        if simplified is True:
            ax.plot(xmin, fmin, markers[i], ls='-', markersize=6, color=colors[i], label=legends[i])
            ax.plot(xmax, fmax, markers[i], ls='-', markersize=6, color=colors[i])

        else:
            ax.plot(xmin, fmin, markers[i], ls='-', markersize=6, color='blue', label='minimum thrust '+legends[i])
            ax.plot(xmax, fmax, markers[i], ls='-', markersize=6, color='red', label='maximum thrust '+legends[i])

        if limit_state:
            x_, y_, _ = intersection_line_line_xy([[xmax[m-1], fmax[m-1]], [xmax[m-2], fmax[m-2]]], [[xmin[n-1], fmin[n-1]], [xmin[n-2], fmin[n-2]]])
            extrapolation_maxy = [fmax[m-1], y_]
            extrapolation_miny = [fmin[n-1], y_]
            extrapolation_maxx = [xmax[m-1], x_]
            extrapolation_minx = [xmin[n-1], x_]
            ax.plot(x_, y_, 'o', ls=' ', markersize=7, color='black')
            ax.plot(extrapolation_maxx, extrapolation_maxy, '', ls='--', color=colors[i])
            ax.plot(extrapolation_minx, extrapolation_miny, '', ls='--', color=colors[i])
            ax.annotate(str(round(max_x/x_, 1)), (x_, y_), textcoords="offset points", xytext=(20, -5), ha='center', size=12)

        if fill:
            xmin_ = append(xmin, x_)
            xmax_ = append(xmax, x_)
            fmin_ = append(fmin, y_)
            fmax_ = append(fmax, y_)

            ax.fill(append(xmin_, xmax_[::-1]), append(fmin_, fmax_[::-1]), color=colors[i], alpha=0.2)
            # ax.fill_between(xmin, fmin, fmax, color=colors[i], alpha=0.2)
            # ax.fill_between(extrapolation_minx, extrapolation_miny, extrapolation_maxy, color=colors[i], alpha=0.2)

    ax1 = plt.axes()
    ax2 = ax1.twiny()
    print([0] + [100*(max_x-tck_x)/(max_x-min_x) for tck_x in ticks_x] + [100])
    print(['1.0'] + [str(round(tck_GSF, 2)) for tck_GSF in ticks_GSF])

    ax2.set_xticks([0] + [100*(max_x-tck_x)/(max_x-min_x) for tck_x in ticks_x] + [100])
    ax2.set_xticklabels(['1.0'] + [str(round(tck_GSF, 2)) for tck_GSF in ticks_GSF] + [''], size=size_axis_data)
    ax1.set_xlabel('thickness/span', size=size_axis_label, weight='bold', labelpad=8)
    ax2.set_xlabel('GSF', size=size_axis_label, weight='bold', labelpad=8)
    ax1.set_ylabel('thrust/weight [%]', size=size_axis_label, weight='bold', labelpad=8)
    ax1.xaxis.set_major_formatter(FormatStrFormatter('%.2f'))
    ax1.yaxis.set_major_formatter(FormatStrFormatter('%.0f'))
    ax1.set_xlim(max_x, min_x)
    ax.set_ylim(min_y, max_y)
    # ax1.set_xticks(arange(max_x, min_x - interval_x, -interval_x))
    ax1.set_yticks(arange(min_y, max_y + interval_y, interval_y))
    ax1.tick_params(axis='both', which='major', labelsize=size_axis_data)

    # plt.axvline(x=(x_ - max_x)/(min_x - max_x)*100, ls='--', color='black') # add    ymin=(max_y-y_)/(max_y-min_y)

    ax.grid(color='silver', linestyle='-', linewidth=0.5)
    if show_legend is True:
        ax.legend(fontsize=size_legend)  # , bbox_to_anchor(0.1, 0.1), ncol=1)

        # ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), fontsize=size_legend)  # , bbox_to_anchor(0.1, 0.1), ncol=1)
        # box = ax.get_position()
        # ax.set_position([box.x0*0.6, box.y0*1.5, box.width * 0.90, box.height*0.90])

    if save:
        plt.savefig(save)
        print('Diagram saved at:', save)

    return plt


def diagram_of_thrust_load_mult(dimension, min_sol, max_sol, limit_state=True, fill=False, xy_limits=None, save=None):
    """ Plot a diagram of Thrusts to the problem of increasing a load multiplier based on the collected data from (n) points.

    Parameters
    ----------
    dimensions : list (n)
        Points with discretised solutions.
    min_sol : list (n)
        Adimensional thrust over weight for minimum thrust.
    max_sol : list (n)
        Adimensional thrust over weight for maximum thrust.
    limit_state : bool, optional
        If yes, it interpolates the limit_state based on the two last solutions.
        The default value is ``True``.
    fill : bool, optional
        If yes, it fills bewteen the curves of min and max.
        The default value is ``False``.
    xy_limits : list of lists, optional
        Freezes the axis of the graphs.
        The default value is ``None`` in which case these limits are calculated automarically.
    save : str, optional
        Path to save the graph.
        The default value is ``False``.

    Returns
    -------
    obj
        Plotter object.

    """

    xmin = xmax = array(dimension)
    fmin = 100.0 * array(min_sol)
    fmax = -100.0 * array(max_sol)
    n = len(xmin)

    size_axis_label = 14
    size_axis_data = 12
    size_legend = 14

    interval_x = abs(dimension[0] - dimension[1])
    interval_y = 10
    if xy_limits:
        [[max_x, min_x], [max_y, min_y]] = xy_limits
    else:
        max_x = max(xmax) + interval_x
        min_x = min(xmax)
        max_y = interval_y - max(fmax) % 10 + max(fmax)
        min_y = min(fmin) - min(fmin) % 10

    fig = plt.figure(figsize=[12, 4])

    ax = fig.add_subplot(1, 1, 1)
    ax.plot(xmin, fmin, 'o', ls='-', markersize=5, color='blue', label='minimum thrust')
    ax.plot(xmax, fmax, 'o', ls='-', markersize=5, color='red', label='maximum thrust')

    if limit_state:
        x_, y_, _ = intersection_line_line_xy([[xmax[n-1], fmax[n-1]], [xmax[n-2], fmax[n-2]]], [[xmin[n-1], fmin[n-1]], [xmin[n-2], fmin[n-2]]])
        extrapolation_max = [fmax[n-1], y_]
        extrapolation_min = [fmin[n-1], y_]
        extrapolation_x = [xmax[n-1], x_]
        ax.plot(x_, y_, 'o', ls=' ', markersize=7, color='black', label='limit state')
        ax.plot(extrapolation_x, extrapolation_max, '', ls='--', color='red')
        ax.plot(extrapolation_x, extrapolation_min, '', ls='--', color='blue')
        ax.annotate(str(round(x_, 1)), (x_, y_), textcoords="offset points", xytext=(0, 10), ha='center')

    if fill:
        ax.fill_between(xmin, fmin, fmax, color='grey', alpha=0.2)
        ax.fill_between(extrapolation_x, extrapolation_min, extrapolation_max, color='grey', alpha=0.2)

    ax1 = plt.axes()
    ax1.set_xlabel('lambda', size=size_axis_label, weight='bold', labelpad=8)
    ax1.set_ylabel('T/W [%]', size=size_axis_label, weight='bold', labelpad=8)
    ax1.xaxis.set_major_formatter(FormatStrFormatter('%.2f'))
    ax1.yaxis.set_major_formatter(FormatStrFormatter('%.0f'))
    ax1.set_xlim(min_x, max_x)
    ax1.set_ylim(min_y, max_y)
    ax1.set_xticks(arange(min_x, max_x + interval_x, interval_x))
    ax1.set_yticks(arange(min_y, max_y + interval_y, interval_y))
    ax1.tick_params(axis='both', which='major', labelsize=size_axis_data)
    box = ax.get_position()
    ax.set_position([box.x0*0.6, box.y0*1.5, box.width * 0.90, box.height*0.90])
    ax.grid(color='silver', linestyle='-', linewidth=0.5)
    ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), fontsize=size_legend)  # , bbox_to_anchor(0.1, 0.1), ncol=1)

    if save:
        plt.savefig(save)

    return plt


def surface_GSF_load_mult(sizes, mins, maxs, legends, save=None):
    """ Plot a diagram of Thrusts based on the collected data on (m) problems each with (n) points.

    Parameters
    ----------
    sizes : list of lists [m-n]
        Points with discretised solutions.
    mins : list of lists [m-n]
        Adimensional thrust over weight for minimum thrust.
    maxs : list of lists [m-n]
        Adimensional thrust over weight for minimum thrust.
    legends : list [m]
        Legend of the (m) problems.
    save : str, optional
        Path to save the graph.
        The default value is ``False``.

    Returns
    -------
    obj
        Plotter object.

    """

    fig = plt.figure()
    ax = fig.gca(projection='3d')

    def flatten_list(li): return [item for sublist in li for item in sublist]
    def flatten_list_inv(li): return [-1*item for sublist in li for item in sublist]

    m = [len(sublist) for sublist in sizes]
    xs = array(flatten_list(sizes)+flatten_list(sizes))
    ys = array(flatten_list(mins)+flatten_list_inv(maxs))
    z_list = flatten_list([[legends[i]]*m[i] for i in range(len(legends))])
    zs = array(z_list + z_list)

    ax.plot_trisurf(xs, ys, zs, cmap=cm.get_cmap('coolwarm'))
    ax.set_xlabel('t/R')
    ax.set_ylabel('T/W')
    ax.set_zlabel('px')

    if save:
        plt.savefig(save)

    return plt


def save_csv_row(thicknesses, solutions, limit_state=True, path=None, title=None):
    """Save a CSV file from the routine with several min/max thrust computations

    Parameters
    ----------
    thicknesses : list
        The list of the thickness computed for min/max thrust
    solutions : [list, list]
        The list of solutions of min/max solutions
    limit_state : bool, optional
        If limit state is included, by default True
    path : str, optional
        Full path to save the file, by default None
    title : str, optional
        Title for the analysis, by default None

    Returns
    -------
    None
        File is saved
    """

    xmin = array(thicknesses[0])
    xmax = array(thicknesses[1])
    fmin = 100.0 * array(solutions[0])
    fmax = -100.0 * array(solutions[1])
    n = len(xmin)
    m = len(xmax)

    if limit_state:
        x_, y_, _ = intersection_line_line_xy([[xmax[m-1], fmax[m-1]], [xmax[m-2], fmax[m-2]]], [[xmin[n-1], fmin[n-1]], [xmin[n-2], fmin[n-2]]])
        # print('Calculated Limit State: (x, y):', x_, y_)
        xmin = append(xmin, x_)
        xmax = append(xmax, x_)
        fmin = append(fmin, y_)
        fmax = append(fmax, y_)

    if path is None:
        path = os.path.join(compas_tno.get('/csv/'), 'test.csv')

    with open(path, mode='w') as csv_file:
        csv_writer = csv.writer(csv_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        if title:
            csv_writer.writerow([title])
        csv_writer.writerow(['X-Min'] + list(xmin))
        csv_writer.writerow(['X-Max'] + list(xmax))
        csv_writer.writerow(['min T/W'] + list(fmin))
        csv_writer.writerow(['max T/W'] + list(fmax))

    return


def open_csv_row(path, cut_last=True, printout=True):
    """Open a CSV file from the routine with several min/max thrust computations

    Parameters
    ----------
    path : str, optional
        Full path to save the file, by default None
    cut_last : bool, optional
        If should cut last element, by default True
    printout : bool, optional
        if should print in the screen, by default True

    Returns
    -------
    thicknesses : list
        The list of the thickness computed for min/max thrust
    solutions : [list, list]
        The list of solutions of min/max solutions
    """

    xmin = []
    xmax = []
    min_thrust = []
    max_thrust = []

    with open(path) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        line_count = 0
        if cut_last is False:
            for row in csv_reader:
                if line_count == 0:
                    if printout:
                        print(f'Title-notcut: {", ".join(row)}')
                elif line_count == 1:
                    xmin.append(row[1:])
                elif line_count == 2:
                    xmax.append(row[1:])
                elif line_count == 3:
                    min_thrust.append(row[1:])
                elif line_count == 4:
                    max_thrust.append(row[1:])
                line_count += 1
        else:
            for row in csv_reader:
                if line_count == 0:
                    if printout:
                        print(f'Title: {", ".join(row)}')
                elif line_count == 1:
                    cut_length = len(row[1:])
                    xmin.append(row[1:cut_length])
                elif line_count == 2:
                    cut_length = len(row[1:])
                    xmax.append(row[1:cut_length])
                elif line_count == 3:
                    cut_length = len(row[1:])
                    min_thrust.append(row[1:cut_length])
                elif line_count == 4:
                    cut_length = len(row[1:])
                    max_thrust.append(row[1:cut_length])
                line_count += 1

    xmin = [float(i) for i in xmin[0]]
    xmax = [float(i) for i in xmax[0]]
    min_thrust = [float(i)/100 for i in min_thrust[0]]
    max_thrust = [-1 * float(i)/100 for i in max_thrust[0]]

    return [xmin, xmax], [min_thrust, max_thrust]


def interpolate_min_thk(sizes, solutions):
    """Interpolate min thickness from sizes and solutions computed

    Parameters
    ----------
    sizes : list
        Sizes computed
    solutions : list
        SOlutions of min/max optimisation

    Returns
    -------
    x_
        Point to add to the Stability domain with the interpolated value
    """

    xmin = array(sizes[0])
    xmax = array(sizes[1])
    fmin = 100.0 * array(solutions[0])
    fmax = -100.0 * array(solutions[1])
    n = len(xmin)
    m = len(xmax)
    x_, _, _ = intersection_line_line_xy([[xmax[m-1], fmax[m-1]], [xmax[m-2], fmax[m-2]]], [[xmin[n-1], fmin[n-1]], [xmin[n-2], fmin[n-2]]])

    return x_


def lookup_folder(folder):
    """Lookup files in a folder

    Parameters
    ----------
    folder : str
        Path to the folder

    Returns
    -------
    files_dict
        Dictionary with the information about the files
    """

    files = os.listdir(folder)
    # print(files)
    jsons = []
    for f in files:
        extension = f.split('.')[-1]
        if extension == 'json':
            jsons.append('.'.join(f.split('.')[:-1]))

    files_dict = {}

    for title in jsons:

        title_split = title.split('_')
        thk = float(title_split[-1])
        type_structure = '_'.join(title_split[:2])
        type_formdiagram = title_split[2]

        if type_formdiagram == 'cross' or type_formdiagram == 'fan':
            type_formdiagram = type_formdiagram + '_fd'

        if 'max' in title_split:
            objective = 'max'
        elif 'min' in title_split:
            objective = 'min'
        elif 'lp' in title_split:
            objective = 'lp'
        elif 'bestfit' in title_split:
            objective = 'bestfit'
        else:
            objective = 'other'
        if 'sag' in title:
            sag = int(title.split('sag_')[-1].split('_')[0])
        else:
            sag = False
        if 'smooth' in title:
            smooth = True
        else:
            smooth = False

        discr_sag_smooth = title_split[title_split.index('discr')+1]

        if not sag and not smooth:
            discr = int(discr_sag_smooth)
        elif sag:
            discr = int(discr_sag_smooth.split('sag')[0])
        else:
            discr = int(discr_sag_smooth.split('smooth')[0])

        data_file = {
            'thk': thk,
            'discretisation': discr,
            'type_structure': type_structure,
            'type_formdiagram': type_formdiagram,
            'objective': objective,
            'sag': sag,
            'smooth': smooth,
        }

        files_dict[title] = data_file

    return files_dict


def filter_min_thk(files_dict, filters=None):
    """Filter structure with minimum thickness in a folder

    Parameters
    ----------
    files_dict : dict
        Dictionary of all files
    filters : dict, optional
        Filters applied, by default None

    Returns
    -------
    limit_form_min, limit_form_max
        The limit values
    """

    limit_thk_min = 100
    limit_thk_max = 100
    limit_form_min = None
    limit_form_max = None

    for key, values in files_dict.items():
        proceed = True
        if filters:
            for filter in filters:
                if filters[filter] == values[filter]:
                    pass
                else:
                    proceed = False
                    break
        if proceed:
            if values['objective'] == 'min':
                if values['thk'] < limit_thk_min:
                    limit_thk_min = values['thk']
                    limit_form_min = {key: values}
            if values['objective'] == 'max':
                if values['thk'] < limit_thk_max:
                    limit_thk_max = values['thk']
                    limit_form_max = {key: values}

    return limit_form_min, limit_form_max


def save_pointcloud(points_lb, points_ub, json_path):
    """Save pointcloud to a JSON file

    Parameters
    ----------
    points_lb : list
        List of points of the intrados
    points_ub : list
        List of points of the extrados
    json_path : str
        Path to save the file
    """

    data = {'UB': {}, 'LB': {}}

    i = 0
    for pt in points_lb:
        data['LB'][i] = pt
        i += 1

    print('Found {0} points in LB'.format(i))

    i = 0
    for pt in points_ub:
        data['UB'][i] = pt
        i += 1

    print('Found {0} points in UB'.format(i))

    with open(json_path, 'w') as outfile:
        json.dump(data, outfile)

    return


# def prune_data(sizes, solutions):

#     xmin = sizes[0]
#     xmax = sizes[1]
#     fmin = solutions[0]
#     fmax = solutions[1]
#     n = len(xmin)
#     m = len(xmax)

#     xmax_ = []
#     xmin_ = []
#     fmax_ = []
#     fmin_ = []

#     if n == m:
#         return sizes, solutions
#     elif n > m and xmin[-1] == xmax[-1]:
#         xmax_ = xmax
#         fmax_ = fmax
#         j = 0
#         for i in range(m):
#             while xmax[i] != xmin[j]:
#                 j = j+1
#             if xmax[i] == xmin[j]:
#                 xmin_.append(xmin[j])
#                 fmin_.append(fmin[j])
#                 j = j+1
#         return [xmin_, xmax_], [fmin_, fmax_]
#     elif n < m and xmin[-1] == xmax[-1]:
#         xmin_ = xmin
#         fmin_ = fmin
#         j = 0
#         for i in range(n):
#             while xmin[i] != xmax[j]:
#                 j = j+1
#             if xmin[i] == xmax[j]:
#                 xmax_.append(xmax[j])
#                 fmax_.append(fmax[j])
#         return [xmin_, xmax_], [fmin_, fmax_]
#     else:
#         print('Last optimisation not with same thickness. Can not prune')
#         print([xmin, xmax], [fmin, fmax])
#         if m > n:
#             diff = m - n
#             print(diff)
#             for i in range(diff):
#                 xmax.pop(m-3)
#                 fmax.pop(m-3)
#             print([xmin, xmax], [fmin, fmax])
#             return [xmin, xmax], [fmin, fmax]
#         else:
#             diff = n - m
#             print(diff)
#             for i in range(diff):
#                 xmin.pop(m-3)
#                 fmin.pop(m-3)
#             print([xmin, xmax], [fmin, fmax])
#             return [xmin, xmax], [fmin, fmax]
#     return sizes, solutions



from scipy.optimize import fmin_slsqp
from scipy.optimize import shgo

from compas.numerical import devo_numpy
from compas.numerical import ga

from .post_process import post_process_general

import time


def run_optimisation_scipy(analysis):
    """ Run nonlinear optimisation problem with SciPy.

    Parameters
    ----------
    analysis : Analysis
        Analysis object with information about optimiser, form and shape.

    Returns
    -------
    analysis : Analysis
        Analysis object optimised.

    """

    optimiser = analysis.optimiser
    solver = optimiser.settings['solver']
    fobj = optimiser.fobj
    fconstr = optimiser.fconstr
    fgrad = optimiser.fgrad
    fjac = optimiser.fjac
    args = [optimiser.M]
    bounds = optimiser.bounds
    x0 = optimiser.x0
    printout = optimiser.settings.get('printout', True)
    grad_choice = optimiser.settings.get('gradient', False)
    jac_choice = optimiser.settings.get('jacobian', False)
    max_iter = optimiser.settings.get('max_iter', 500)
    callback = optimiser.callback

    if grad_choice is False:
        fgrad = None
    if jac_choice is False:
        fjac = None

    start_time = time.time()

    if solver == 'slsqp' or solver == 'SLSQP':
        fopt, xopt, exitflag, niter, message = _slsqp(fobj, x0, bounds, fgrad, fjac, printout, fconstr, args, max_iter, callback)
    elif solver == 'shgo':
        dict_constr = []
        for i in range(len(fconstr(x0, *args))):
            args_constr = list(args)
            args_constr.append(i)
            args_constr.append(fconstr)
            dict_ = {
                'type': 'ineq',
                'fun': _shgo_constraint_wrapper,
                'args': args_constr,
            }
            dict_constr.append(dict_)
        args_constr[len(args_constr)-2] = 0
        result = _shgo(fobj, bounds, True, dict_constr, args)
        fopt = result['fun']
        xopt = result['x']
        sucess = result['success']
        message = result['message']
        if sucess is True:
            exitflag = 0
        else:
            print(message)

    elapsed_time = time.time() - start_time
    if printout:
        print('Solving Time: {0:.1f} sec'.format(elapsed_time))

    # Store output info in optimiser

    optimiser.exitflag = exitflag
    optimiser.time = elapsed_time
    optimiser.fopt = float(fopt)
    optimiser.xopt = xopt
    optimiser.niter = niter
    optimiser.message = message

    post_process_general(analysis)

    return analysis


def _slsqp(fn, qid0, bounds, fprime, fprime_ieqcons, printout, fieq, args, iter, callback):
    pout = 2 if printout else 0
    opt = fmin_slsqp(fn, qid0, args=args, disp=pout, fprime=fprime, f_ieqcons=fieq, fprime_ieqcons=fprime_ieqcons, bounds=bounds, full_output=1, iter=iter, callback=callback)

    return opt[1], opt[0], opt[3], opt[2], opt[4]


def _shgo(fn, bounds, printout, dict_constr, args):

    res = shgo(fn, bounds, args=args, constraints=dict_constr, n=3, iters=3, options={'disp': True})

    return res


def _shgo_constraint_wrapper(x, *args_constr):
    length = len(args_constr)
    args = args_constr[:length-2]
    i = args_constr[length-2]
    fconstr = args_constr[length-1]
    gi = fconstr(x, *args)[i]

    return gi


def _cobyla(fn, qid0, bounds, printout, fieq, args):

    # pout = 2 if printout else 0
    # opt  = fmin_cobyla(fn, qid0, args=args, disp=pout, bounds=bounds, full_output=1, iter=500, f_ieqcons=fieq)

    # W.I.P

    return None


def _diff_evo(fn, bounds, population, generations, printout, plot, frange, args):

    return devo_numpy(fn=fn, bounds=bounds, population=population, generations=generations, printout=printout,
                      plot=plot, frange=frange, args=args)


def _ga(fn, fit_type, num_var, boundaries, num_gen, num_pop, args):

    return ga(fit_function=fn, fit_type=fit_type, num_var=num_var, boundaries=boundaries, num_gen=num_gen, num_pop=num_pop, fargs=args)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from copy import deepcopy
from compas.geometry import subtract_vectors
from compas.geometry import length_vector
from compas.geometry import cross_vectors
from compas.geometry import normalize_vector

from compas_tno.shapes import MeshDos

from compas.datastructures import Datastructure

import math

__all__ = ['Shape']


class Shape(Datastructure):
    """The ``Shape`` class deals with the geometry of a masonry vault, where the definition of Extrados, Intrados and Middle surfaces are of interest.

    The class imports the attributes and set ups from ``compas_tna.diagrams.FormDiagram`` and include some
    functionalities useful for the assessment of masonry structures. Most relevant edge and vertex attributes
    are listed below:

    Parameters
    ----------
    intrados : MeshDos
        Mesh representing the intrados of the masonry
    extrados : MeshDos
        Mesh representing thhe extrados of the masonry
    middle : MeshDos
        Mesh representing the midde surface of the masonry
    name : str
        The name of the shape
    total_selfweight : float
        Total weight of the masonry in kN.
    area : float
        Area of the middle surface of the masonry
    ro : float
        The density of the masonry
    fill : bool
        Whether or not a fill should be considered up to a given elevation
    fill_ro : float
        The density of the masonry in the fill

    Attributes
    ----------
    type : str
        The type of the masory. Types include:
        ['crossvault', 'pavillionvault', 'dome', 'dome_polar', 'dome_spr', 'parabolic_shell', 'arch', 'pointed_arch', 'pointed_crossvault']
    thk : float
        The thickness of the masonry
    discretisation : int
        The discretisation level
    xy_span : list
        The limits of the xy box (for rectangular form diagram)
    t : float
        The parameter assumed when no intrados projection is found

    Examples
    --------
    >>> from compas_tno.shapes import Shape
    >>> data = {'type': 'crossvault', 'thk': 0.5, 'discretisation': 10, 'xy_span': [[0.0, 10.0], [0.0, 10.0]], 't': 0.0 }
    >>> shape = Shape.from_library(data)


    Notes
    -----
    A ``Shape`` has the following constructor functions

    *   ``from_library`` : Construct the shape from a dictionary with instructions, the library supports the creation of parametric arches, domes, and vaults.
    *   ``from_rhinomesh`` : Construct Extrados, Intrados and Middle surfaces from RhinoMeshes.
    *   ``from_rhinosurface`` : Construct Extrados, Intrados and Middle surfaces using the U and V isolines.

    A ``Shape`` has the following elements:

    *   ``data``

        *   ``type``  : The type of the vault to be constructed.
        *   ``xy_span``  : Planar range of the structure.
        *   ``discretisation``  : Density of the highfield.
        *   ``thickness`` : Mean thickness of the vault.
        *    ``t`` : The numerical for the negative bounds on the height of the fixed nodes.

    *   ``discretisation``

        *   ``array``    : (2 x n) array with the coordinate of the n points with intrados, extrados and middle evaluated.

    *   ``intrados``

        *   ``Mesh``    : Mesh representing intrados.

    *   ``extrados``

        *   ``Mesh``    : Mesh representing extrados.

    *   ``middle``

        *   ``Mesh``    : Mesh representing middle surface

    """

    # __module__ = 'compas_tno.shapes'

    def __init__(self):
        super(Shape, self).__init__()
        self.datashape = {
            'type': None,
            'thk': 1.0,
            'discretisation': [20, 20],
            'xy_span': [0.0, 10.0],
            't': 10.0,
        }
        self.intrados = None
        self.extrados = None
        self.middle = None
        self.name = "shape"
        self.total_selfweight = 0.0
        self.area = 0.0
        self.ro = 20.0
        self.fill = False
        self.fill_ro = 14.0

    @property
    def data(self):
        """ A data dict representing the shape data structure for serialization.
        """
        dataintrados = None
        dataextrados = None
        datamiddle = None

        if self.intrados:
            dataintrados = self.intrados.to_data()
        if self.extrados:
            dataextrados = self.extrados.to_data()
        if self.middle:
            datamiddle = self.middle.to_data()

        data = {
            'datashape': self.datashape,
            'intrados': dataintrados,
            'extrados': dataextrados,
            'middle': datamiddle,
            'name': self.name,
            'total_selfweight': self.total_selfweight,
            'area': self.area,
            'ro': self.ro,
            'fill': self.fill,
            'fill_ro': self.fill_ro
        }
        return data

    @data.setter
    def data(self, data):
        if 'data' in data:
            data = data['data']
        self.datashape = data.get('datashape') or {}

        self.name = data.get('name') or "shape"
        self.total_selfweight = data.get('total_selfweight') or 0.0
        self.area = data.get('area') or 0.0
        self.ro = data.get('ro') or 20.0
        self.fill = data.get('fill') or False
        self.fill_ro = data.get('ro') or 20.0

        dataintrados = data.get('intrados')
        dataextrados = data.get('extrados')
        datamiddle = data.get('middle')

        self.intrados = MeshDos.from_data(dataintrados) if dataintrados else None
        self.extrados = MeshDos.from_data(dataextrados) if dataintrados else None
        self.middle = MeshDos.from_data(datamiddle) if dataintrados else None

    @classmethod
    def from_library(cls, data):
        """Construct a Shape from a library.

        Parameters
        ----------
        data : dict
            Dictionary with the settings to create the vault.

        Returns
        -------
        Shape
            A Shape object.

        """

        from compas_tno.shapes import cross_vault_highfields
        from compas_tno.shapes import pavillion_vault_highfields
        from compas_tno.shapes import set_dome_heighfield
        from compas_tno.shapes import set_dome_with_spr
        from compas_tno.shapes import set_dome_polar_coord
        from compas_tno.shapes import arch_shape
        from compas_tno.shapes import arch_shape_polar
        from compas_tno.shapes import pointed_arch_shape
        from compas_tno.shapes import pointed_vault_heightfields
        from compas_tno.shapes import domical_vault
        from compas_tno.shapes import parabolic_shell_highfields

        shape = cls()

        typevault = data['type']
        thk = data['thk']
        discretisation = data['discretisation']
        t = data.get('t', 0.0)
        expanded = data.get('expanded', False)

        if typevault == 'crossvault':
            xy_span = data['xy_span']
            intrados, extrados, middle = cross_vault_highfields(xy_span, thk=thk, discretisation=discretisation, t=t, expanded=expanded)
        elif typevault == 'pavillionvault':
            xy_span = data['xy_span']
            spr_angle = data.get('spr_angle', 0.0)
            intrados, extrados, middle = pavillion_vault_highfields(xy_span, thk=thk, discretisation=discretisation, t=t, spr_angle=spr_angle, expanded=expanded)
        elif typevault == 'dome':
            center = data['center']
            radius = data['radius']
            intrados, extrados, middle = set_dome_heighfield(center, radius=radius, thk=thk, discretisation=discretisation, t=t, expanded=expanded)
        elif typevault == 'dome_polar':
            center = data['center']
            radius = data['radius']
            intrados, extrados, middle = set_dome_polar_coord(center, radius=radius, thk=thk, discretisation=discretisation, t=t)
        elif typevault == 'dome_spr':
            center = data['center']
            radius = data['radius']
            theta = data['theta']
            intrados, extrados, middle = set_dome_with_spr(center, radius=radius, thk=thk, theta=theta, discretisation=discretisation, t=t)
        elif typevault == 'parabolic_shell':
            xy_span = data['xy_span']
            intrados, extrados, middle = parabolic_shell_highfields(xy_span, thk=thk, discretisation=discretisation, t=t)
        elif typevault == 'arch':
            H = data['H']
            L = data['L']
            b = data['b']
            x0 = data['x0']
            intrados, extrados, middle = arch_shape(H=H, L=L, x0=x0, thk=thk, total_nodes=discretisation, b=b, t=t)
        elif typevault == 'arch_polar':
            H = data['H']
            L = data['L']
            b = data['b']
            x0 = data['x0']
            intrados, extrados, middle = arch_shape_polar(H=H, L=L, x0=x0, thk=thk, total_nodes=discretisation, b=b)
        elif typevault == 'pointed_arch':
            hc = data['hc']
            L = data['L']
            b = data['b']
            x0 = data['x0']
            intrados, extrados, middle = pointed_arch_shape(hc=hc, L=L, x0=x0, thk=thk, total_nodes=discretisation, b=b, t=t)
        elif typevault == 'pointed_crossvault':
            xy_span = data['xy_span']
            hc = data['hc']
            hm = data.get('hm', None)
            he = data.get('he', None)
            intrados, extrados, middle = pointed_vault_heightfields(xy_span=xy_span, discretisation=discretisation, t=t, hc=hc, he=he, hm=hm, thk=thk)
        elif typevault == 'domicalvault':
            xy_span = data['xy_span']
            center = [sum(xy_span[0])/2, sum(xy_span[1])/2]
            radius = data.get('radius', None)
            intrados, extrados, middle = domical_vault(xy_span=xy_span, thk=thk, radius=radius, center=center, t=t, discretisation=discretisation)

        shape.datashape = data
        shape.intrados = intrados
        shape.extrados = extrados
        shape.middle = middle

        shape.area = middle.area()
        shape.volume = shape.compute_volume()
        shape.total_selfweight = shape.compute_selfweight()

        return shape

    @classmethod
    def from_library_proxy(cls, data):
        """Construct a Shape from a library using a proxy implementation (compatible with Rhinoceros).

        Parameters
        ----------
        data : dict
            Dictionary with the setttings to create the vault.

        Returns
        -------
        Shape
            A Shape object.

        """

        from compas.rpc import Proxy

        proxy = Proxy()
        proxy.package = 'compas_tno.shapes'

        typevault = data['type']
        thk = data['thk']
        discretisation = data['discretisation']
        t = data.get('t', 0.0)
        # expanded = data.get('expanded', False)

        if typevault == 'crossvault':
            xy_span = data['xy_span']
            intra_data, extra_data, middle_data = proxy.cross_vault_highfields_proxy(xy_span, thk=thk, discretisation=discretisation, t=t)
        elif typevault == 'pavillionvault':
            xy_span = data['xy_span']
            intra_data, extra_data, middle_data = proxy.pavillion_vault_highfields_proxy(xy_span, thk=thk, discretisation=discretisation, t=t)
        elif typevault == 'pointed_crossvault':
            xy_span = data['xy_span']
            hc = data['hc']
            hm = data.get('hm', None)
            he = data.get('he', None)
            intra_data, extra_data, middle_data = proxy.pointed_vault_heightfields_proxy(xy_span=xy_span, thk=thk, discretisation=discretisation, t=t, hc=hc, he=he, hm=hm)
        elif typevault == 'dome':
            center = data['center']
            radius = data['radius']
            intra_data, extra_data, middle_data = proxy.dome_heightfields_proxy(center, radius=radius, thk=thk, discretisation=discretisation, t=t)

        intrados = MeshDos.from_data(intra_data)
        extrados = MeshDos.from_data(extra_data)
        middle = MeshDos.from_data(middle_data)

        shape = cls()
        shape.datashape = data
        shape.intrados = intrados
        shape.extrados = extrados
        shape.middle = middle

        shape.area = middle.area()
        shape.volume = shape.compute_volume()
        shape.total_selfweight = shape.compute_selfweight()

        return shape

    @classmethod
    def create_dome(cls, center=[5.0, 5.0, 0.0], radius=5.0, thk=0.5, discretisation=[16, 40], t=0.0):
        """Create the shape representing a Hemispheric Dome

        Parameters
        ----------
        center : [float, float, float], optional
            Center of the dome, by default [5.0, 5.0, 0.0]
        radius : float, optional
            Central radius of the dome, by default 5.0
        thk : float, optional
            Thickness of the dome, by default 0.5
        discretisation : [float, float], optional
            Discretisation of the parallel and meridians, by default [16, 40]
        t : float, optional
            Negative thickness to consider if the intrados can not be interpolated in the point, by default 0.5

        Returns
        -------
        shape : Shape
            The shape of the dome.

        """

        data = {'type': 'dome', 'thk': thk, 'discretisation': discretisation, 'center': center, 'radius': radius, 't': t}

        return cls().from_library(data)

    @classmethod
    def create_dome_polar(cls, center=[5.0, 5.0, 0.0], radius=5.0, thk=0.5, discretisation=[16, 40], t=0.0):
        """Create the shape representing a Hemispheric Dome

        Parameters
        ----------
        center : [float, float, float], optional
            Center of the dome, by default [5.0, 5.0, 0.0]
        radius : float, optional
            Central radius of the dome, by default 5.0
        thk : float, optional
            Thickness of the dome, by default 0.5
        discretisation : [float, float], optional
            Discretisation of the parallel and meridians, by default [16, 40]
        t : float, optional
            Negative thickness to consider if the intrados can not be interpolated in the point, by default 0.0

        Returns
        -------
        shape : Shape
            The shape of the dome.

        """

        data = {'type': 'dome_polar', 'thk': thk, 'discretisation': discretisation, 'center': center, 'radius': radius, 't': t}

        return cls().from_library(data)

    @classmethod
    def create_arch(cls, H=1.00, L=2.0, x0=0.0, thk=0.20, b=0.5, t=0.5, discretisation=100):
        """Create the shape representing a circular arch.

        Parameters
        ----------
        H : float, optional
            Height of the arch, by default 1.00
        L : float, optional
            Span of the arch, by default 2.0
        x0 : float, optional
            Starting coordinate of the arch , by default 0.0
        thk : float, optional
            Thickness of the arch, by default 0.20
        b : float, optional
            Out-of-plane measure of the arch, by default 0.5
        t : float, optional
            Parameter for lower bound in nodes in the boundary, by default 0.0
        discretisation : int, optional
            Density of the shape, by default 100

        Returns
        -------
        shape : Shape
            The shape of the dome.

        """

        data = {'type': 'arch', 'thk': thk, 'discretisation': discretisation, 'H': H, 'L': L, 'x0': x0, 'b': b, 't': t}

        return cls().from_library(data)

    @classmethod
    def create_arch_polar(cls, H=1.00, L=2.0, x0=0.0, thk=0.20, b=0.5, t=0.0, discretisation=100):
        """Create the shape representing a circular arch.

        Parameters
        ----------
        H : float, optional
            Height of the arch, by default 1.00
        L : float, optional
            Span of the arch, by default 2.0
        x0 : float, optional
            Starting coordinate of the arch , by default 0.0
        thk : float, optional
            Thickness of the arch, by default 0.20
        b : float, optional
            Out-of-plane measure of the arch, by default 0.5
        t : float, optional
            Parameter for lower bound in nodes in the boundary, by default 0.0
        discretisation : int, optional
            Density of the shape, by default 100

        Returns
        -------
        shape : Shape
            The shape of the dome.

        """

        data = {'type': 'arch_polar', 'thk': thk, 'discretisation': discretisation, 'H': H, 'L': L, 'x0': x0, 'b': b, 't': t}

        return cls().from_library(data)

    @classmethod
    def create_crossvault(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], thk=0.50, t=0.0, spr_angle=None, discretisation=[100, 100]):
        """Create the shape representing a Crossvault

        Parameters
        ----------
        xy_span : [list[float], list[float]], optional
            xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
        thk : float, optional
            The thickness of the vault, by default 0.5
        t : float, optional
            Parameter for lower bound in nodes in the boundary, by default 0.0
        spr_angle : float, optional
            Springing angle, by default None
        discretisation : list|int, optional
            Level of discretisation of the shape, by default [100, 100]

        Returns
        -------
        shape : Shape
            The shape of the dome.

        """

        if spr_angle:  # amplify the bounds if springing angle is present
            x0, xf = xy_span[0]
            alpha = 1/math.cos(math.radians(spr_angle))
            Ldiff = (xf - x0) * (alpha - 1)
            xyspan_shape = [[x0 - Ldiff/2, xf + Ldiff/2], [x0 - Ldiff/2, xf + Ldiff/2]]
            xy_span = xyspan_shape

        data = {'type': 'crossvault', 'thk': thk, 'discretisation': discretisation, 'xy_span': xy_span, 't': t}

        return cls().from_library(data)

    @classmethod
    def create_pointedcrossvault(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], thk=0.5, discretisation=[10, 10], hc=8.0, he=None, hm=None,  t=0.0):
        """Create the shape representing a Pointed Crossvault

        Parameters
        ----------
        xy_span : list, optional
            [description], by default [[0.0, 10.0], [0.0, 10.0]]
        thk : float, optional
            [description], by default 0.5
        discretisation : list, optional
            [description], by default [10, 10]
        hc : float, optional
            Height in the middle point of the vault, by default 8.0
        he : [float, float, float, float], optional
            Height of the opening mid-span for each of the quadrants, by default None
        hm : [float, float, float, float], optional
            Height of each quadrant center (spadrel), by default None
        t : float, optional
            Parameter for lower bound in nodes in the boundary, by default 0.0

        Returns
        -------
        shape : Shape
            The shape of the dome.

        """

        data = {'type': 'pointed_crossvault', 'xy_span': xy_span, 'thk': thk, 'discretisation': discretisation, 't': t, 'hc': hc, 'he': he, 'hm': hm}

        return cls().from_library(data)

    @classmethod
    def create_pavillionvault(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], thk=0.50, t=0.0, discretisation=[100, 100], spr_angle=0.0, expanded=False):
        """Create the shape representing a Pavillion Vault

        Parameters
        ----------
        xy_span : [list[float], list[float]], optional
            xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
        thk : float, optional
            The thickness of the vault, by default 0.5
        t : float, optional
            Parameter for lower bound in nodes in the boundary, by default 0.0
        discretisation : list|int, optional
            Level of discretisation of the shape, by default [100, 100]
        spr_angle : float, optional
            Springing angle, by default 0.0
        expanded : bool, optional
            If the extrados should extend beyond the floor plan, by default False

        Returns
        -------
        shape : Shape
            The shape of the dome.

        """

        data = {'type': 'pavillionvault', 'thk': thk, 'discretisation': discretisation, 'xy_span': xy_span, 't': t, 'spr_angle': spr_angle, 'expanded': expanded}

        return cls().from_library(data)

    @classmethod
    def from_meshes(cls, intrados, extrados, middle=None, treat_creases=False, data={'type': 'general', 'thk': 0.50, 't': 0.0}):
        """Construct a Shape from meshes for intrados, extrados, and middle.

        Parameters
        ----------
        intrados : :class:`~compas.datastructures.Mesh`
            Mesh for intrados.
        extrados : :class:`~compas.datastructures.Mesh`
            Mesh for extrados.
        middle : mesh, optional
            Mesh for middle.
            The default value is ``None``, in a case in which no middle surface is assigned
        data : dict, optional
            Dictionary with the data about the structure.
            The default value is a typical general dictionary.

        Returns
        -------
        Shape
            A Shape object.

        """
        shape = cls()

        shape.intrados = MeshDos.from_mesh(intrados)
        shape.extrados = MeshDos.from_mesh(extrados)
        if middle:
            shape.middle = MeshDos.from_mesh(middle)
        else:
            shape.interpolate_middle_from_ub_lb()
            shape.middle = MeshDos.from_mesh(shape.intrados.copy())
            for key in shape.intrados.vertices():
                ub = shape.extrados.vertex_attribute(key, 'z')
                lb = shape.intrados.vertex_attribute(key, 'z')
                shape.middle.vertex_attribute(key, 'z', (ub + lb)/2)

        if data:
            shape.datashape = data

        return shape

    @classmethod
    def from_pointcloud(cls, intrados_pts, extrados_pts, middle_pts=None, data={'type': 'general', 't': 0.0}):
        """Construct a Shape from a pointcloud.

        Parameters
        ----------
        intrados_pts : list
            List of points collected in the intrados.
        extrados_pts : list
            List of points collected in the extrados.
        middle_pts : list, optional
            List of points collected in the middle.
            The default value is ``None``, in which case no middle surface is connstructed
        data : dict (None)
            Dictionary with the data in required.

        Returns
        -------
        Shape
            A Shape object.

        """

        from compas_tno.utilities import mesh_from_pointcloud

        mesh_extrados = mesh_from_pointcloud(extrados_pts)
        mesh_intrados = mesh_from_pointcloud(intrados_pts)
        mesh_middle = None
        if middle_pts:
            mesh_middle = mesh_from_pointcloud(middle_pts)
        shape = cls().from_meshes(mesh_intrados, mesh_extrados, middle=mesh_middle, data=data)

        return shape

    @classmethod
    def from_middle(cls, middle, thk=0.50, treat_creases=False, printout=False, data={'type': 'general', 't': 0.0, 'xy_span': [[0.0, 10.0], [0.0, 10.0]]}):
        """Construct a Shape from a middle mesh and a thickness (offset happening to compute new extrados).

        Parameters
        ----------
        middle : :class:`~compas.datastructures.Mesh`
            Mesh with middle. Topology will be considered.
        data : dict (None)
            Dictionary with the data in required.

        Returns
        -------
        Shape
            A Shape object.

        """
        middle = MeshDos.from_mesh(middle)
        if treat_creases:
            middle.identify_creases_at_diagonals(xy_span=data['xy_span'])
            middle.store_normals(correct_creases=True)
        else:
            middle.store_normals()
        extrados_mesh = middle.offset_mesh(thk/2, direction='up')
        intrados_mesh = middle.offset_mesh(thk/2, direction='down')
        data['thk'] = thk
        shape = cls().from_meshes(intrados_mesh, extrados_mesh, middle=middle, data=data)

        return shape

    @classmethod
    def from_intrados(cls, intrados, thk=0.50, treat_creases=False, printout=False, data={'type': 'general', 't': 0.0, 'xy_span': [[0.0, 10.0], [0.0, 10.0]]}):
        """Construct a Shape from an intrados mesh and a thickness (offset happening to both sides).

        Parameters
        ----------
        intrados : :class:`~compas.datastructures.Mesh`
            Mesh with middle. Topology will be considered.
        data : dict (None)
            Dictionary with the data in required.

        Returns
        -------
        Shape
            A Shape object.

        """
        intrados = MeshDos.from_mesh(intrados)
        if treat_creases:
            intrados.identify_creases_at_diagonals(xy_span=data['xy_span'])
            intrados.store_normals(correct_creases=True)
        else:
            intrados.store_normals()
        extrados = intrados.offset_mesh(thk, direction='up')

        data['thk'] = thk
        # bbox = intrados.bounding_box_xy()
        # data['xy_span'] = TAKE THIS FROM BBOX

        shape = cls().from_meshes(intrados, extrados, data=data)

        return shape

    # @classmethod
    # def from_middle_pointcloud(cls, middle_pts, topology=None, thk=0.50, treat_creases=False, printout=False,
    #                            data={'type': 'general', 't': 0.0, 'xy_span': [[0.0, 10.0], [0.0, 10.0]]}):
    #     """Construct a Shape from a pointcloud.

    #     Parameters
    #     ----------
    #     middle_pts : list
    #         List of points collected in the intrados.
    #     topology : mesh (None)
    #         If a mesh is given the middle, intrados and extrados meshes will repeat the topology.
    #     data : dict (None)
    #         Dictionary with the data in required.

    #     Returns
    #     -------
    #     Shape
    #         A Shape object.

    #     """

    #     if not topology:
    #         middle = MeshDos.from_points_delaunay(middle_pts)
    #     else:
    #         middle = MeshDos.from_topology_and_pointcloud(topology, array(middle_pts))
    #     if treat_creases:
    #         middle.identify_creases_at_diagonals(xy_span=data['xy_span'])
    #         middle.store_normals(correct_creases=True)
    #     else:
    #         middle.store_normals()
    #     extrados_mesh = middle.offset_mesh(thk/2, direction='up')
    #     intrados_mesh = middle.offset_mesh(thk/2, direction='down')
    #     data['thk'] = thk
    #     shape = cls().from_meshes(intrados_mesh, extrados_mesh, middle=middle, data=data)

    #     return shape

    @classmethod
    def from_pointcloud_and_topology(cls, meshtopology, intrados_pts, extrados_pts, middle_pts=None, fill_pts=None, data={'type': 'general', 't': 0.0}):
        """Construct a Shape from a pointcloud and a given topology.

        Parameters
        ----------
        meshtopology: mesh
            Mesh with the topology to be used.
        intrados_pts : list
            List of points collected in the intrados.
        extrados_pts : list
            List of points collected in the extrados.
        middle_pts : list, optional
            List of points collected in the middle.
            The default value is ``None``, in which case no middle surface is constructed
        fill_pts : list, optional
            List of points collected in the fill.
            The default value is ``None``, in which case no fill surface is constructed
        data : dict, optional
            Dictionary with the data in required. The default is None.

        Returns
        -------
        Shape
            A Shape object.

        """

        from compas_tno.utilities import create_mesh_from_topology_and_pointcloud

        intrados_mesh = create_mesh_from_topology_and_pointcloud(meshtopology, intrados_pts)
        extrados_mesh = create_mesh_from_topology_and_pointcloud(meshtopology, extrados_pts)
        middle_mesh = None
        if middle_pts:
            middle_mesh = create_mesh_from_topology_and_pointcloud(meshtopology, middle_pts)
        shape = cls().from_meshes(intrados_mesh, extrados_mesh, middle=middle_mesh, data=data)
        if fill_pts:
            fill_mesh = create_mesh_from_topology_and_pointcloud(meshtopology, fill_pts)
            shape.fill = fill_mesh

        return shape

    @classmethod
    def from_meshes_and_formdiagram(cls, form, intrados, extrados, middle=None, data={'type': 'general', 't': 0.0}):
        """Construct a Shape from a pair of meshes and a formdiagram that will have its topology copied and normals copied.

        Parameters
        ----------
        form: FormDiagram
            Form Diagram with the topology to be used
        intrados : mesh
            Mesh for intrados
        extrados : mesh
            Mesh for extrados
        middle : mesh, optional
            Mesh for middle, the default is None
        data : dict,  optional
            Dictionary with the data in required, the default is None

        Returns
        -------
        Shape
            A Shape object.

        """

        intrados_mesh = MeshDos.from_topology_and_mesh(form, intrados, keep_normals=True)
        extrados_mesh = MeshDos.from_topology_and_mesh(form, extrados, keep_normals=True)
        shape = cls().from_meshes(intrados_mesh, extrados_mesh, middle=middle, data=data)

        return shape

    @classmethod
    def from_formdiagram_and_attributes(cls, form, data={'type': 'general', 't': 0.0}):
        """Construct a Shape from the form diagram and its attributes 'ub' and 'lb'.

        Parameters
        ----------
        form: FormDiagram
            Form Diagram with the topology to be used.
        intrados : mesh
            Mesh for intrados.
        extrados : mesh
            Mesh for extrados.
        middle : mesh, optional
            Mesh for middle, the default is None
        data : dict, optional
            Dictionary with the data if required, the default is None

        Returns
        -------
        Shape
            A Shape object.

        """

        vertices, faces = form.to_vertices_and_faces()
        intra = MeshDos.from_vertices_and_faces(vertices, faces)
        extra = MeshDos.from_vertices_and_faces(vertices, faces)
        middle = MeshDos.from_vertices_and_faces(vertices, faces)

        for key in form.vertices():
            intra.vertex_attribute(key, 'z', form.vertex_attribute(key, 'lb'))
            extra.vertex_attribute(key, 'z', form.vertex_attribute(key, 'ub'))
            middle.vertex_attribute(key, 'z', form.vertex_attribute(key, 'target'))

        shape = cls().from_meshes(intra, extra, middle=middle, data=data)

        return shape

    @classmethod
    def from_assembly(self):
        """Create a TNO Shape from an assembly

        Returns
        -------
        Shape
            The shape object
        """

        raise NotImplementedError

    def interpolate_middle_from_ub_lb(self, intrados=None, extrados=None):
        """Interpolate the middle surface based on intrados and extrados

        Parameters
        ----------
        intrados : Mesh, optional
            Intrados to consider, by default None
        extrados : Mesh, optional
            Extrados to consider, by default None

        Returns
        -------
        None
            Middle surface is added to the Shape.
        """

        if not intrados:
            intrados = self.intrados

        if not extrados:
            extrados = self.extrados

        self.middle = MeshDos.from_mesh(self.intrados.copy())

        for key in self.intrados.vertices():
            ub = self.extrados.vertex_attribute(key, 'z')
            lb = self.intrados.vertex_attribute(key, 'z')
            self.middle.vertex_attribute(key, 'z', (ub + lb)/2)

        return

    def store_normals(self, mark_fixed_LB=True, plot=False):
        """Store the normals of the shape

        Parameters
        ----------
        mark_fixed_LB : bool, optional
            If vertices in the boundary are taken specially, by default True
        plot : bool, optional
            If plots should appear, by default False

        Returns
        -------
        None
            Normals are added as attributes.
        """

        intrados = self.intrados
        extrados = self.extrados

        intrados.store_normals()
        extrados.store_normals()

        intrados.vertices_attribute('is_outside', False)
        extrados.vertices_attribute('is_outside', False)

        if mark_fixed_LB:  # Look for the vertices with LB == 't'
            try:
                t = self.datashape['t']
            except BaseException:
                print('No t is assigned to vertices in intrados.')
                t = None
            if t is not None:
                for key in intrados.vertices():
                    if abs(intrados.vertex_attribute(key, 'z') - t) < 10e-3:
                        intrados.vertex_attribute(key, 'is_outside', True)
                    if abs(extrados.vertex_attribute(key, 'z') - t) < 10e-3:
                        extrados.vertex_attribute(key, 'is_outside', True)

        if plot:
            intrados.plot_normals()
            extrados.plot_normals()

        return

    def analytical_normals(self, assume_shape=None, mark_fixed_LB=True):
        """Compute normals analytically for some shapes

        Parameters
        ----------
        assume_shape : dict, optional
            A dictionary with data of the shape to assume, by default None
        mark_fixed_LB : bool, optional
            If vertices in the boundary should be considered specially, by default True

        Returns
        -------
        None
            Normals added as attribute.
        """

        if assume_shape:
            data = assume_shape
        else:
            data = self.datashape

        if data['type'] == 'dome':
            xc, yc = data['center']
        else:
            return Exception

        intrados = self.intrados
        extrados = self.extrados

        for key in intrados.vertices():
            x, y, zlb = intrados.vertex_coordinates(key)
            _, _, zub = extrados.vertex_coordinates(key)
            nlb = normalize_vector([(x - xc), (y - yc), zlb])
            nub = normalize_vector([(x - xc), (y - yc), zub])
            intrados.vertex_attribute(key, 'n', nlb)
            extrados.vertex_attribute(key, 'n', nub)

        intrados.vertices_attribute('is_outside', False)
        extrados.vertices_attribute('is_outside', False)

        if mark_fixed_LB:  # Look for the vertices with LB == 't'
            try:
                t = self.datashape['t']
            except BaseException:
                print('No t is assigned to vertices in intrados.')
                t = None
            if t is not None:
                for key in intrados.vertices():
                    if abs(intrados.vertex_attribute(key, 'z') - t) < 10e-3:
                        intrados.vertex_attribute(key, 'is_outside', True)
                    if abs(extrados.vertex_attribute(key, 'z') - t) < 10e-3:
                        extrados.vertex_attribute(key, 'is_outside', True)

        return

    def compute_selfweight(self):
        """Compute and returns the total selfweight of the structure based on the area and thickness in the data.

        Returns
        -------
        total_selfweight : float
            The selfweight.
        """

        middle = self.middle
        ro = self.ro
        try:
            thk = self.datashape['thk']
            area = middle.area()
            total_selfweight = thk * area * ro
        except BaseException:
            intrados = self.intrados
            extrados = self.extrados
            total_selfweight = 0
            for key in intrados.vertices():
                h = extrados.vertex_attribute(key, 'z') - intrados.vertex_attribute(key, 'z')
                vol = h * middle.vertex_projected_area(key)
                total_selfweight += vol * ro

        return total_selfweight

    def compute_volume(self):
        """Compute and returns the vollume of the structure based on the area and thickness in the data.

        Returns
        -------
        volume : float
            The volume.
        """

        middle = self.middle
        thk = self.datashape['thk']
        area = middle.area()
        volume = thk * area

        return volume

    def add_fill_with_height(self, height, fill_ro=20.0):
        """Compute the weight of the fill applied based on the height.

        Parameters
        ----------
        height : float
            The height until which infill should be observed.
        fill_ro : float, optional
            The density of the infill, the default is 20.0


        """

        self.extrados_fill = deepcopy(self.extrados)
        volume = 0.
        proj_area_total = 0.

        for key in self.extrados_fill.vertices():
            zi = self.extrados_fill.vertex_attribute(key, 'z')
            if zi < height:
                self.extrados_fill.vertex_attribute(key, 'z', height)
                # this should become a method
                area = 0.
                p0 = self.extrados_fill.vertex_coordinates(key)
                p0[2] = 0
                for nbr in self.extrados_fill.halfedge[key]:
                    p1 = self.extrados_fill.vertex_coordinates(nbr)
                    p1[2] = 0
                    v1 = subtract_vectors(p1, p0)
                    fkey = self.extrados_fill.halfedge[key][nbr]
                    if fkey is not None:
                        p2 = self.extrados_fill.face_centroid(fkey)
                        p2[2] = 0
                        v2 = subtract_vectors(p2, p0)
                        area += length_vector(cross_vectors(v1, v2))
                    fkey = self.extrados_fill.halfedge[nbr][key]
                    if fkey is not None:
                        p3 = self.extrados_fill.face_centroid(fkey)
                        p3[2] = 0
                        v3 = subtract_vectors(p3, p0)
                        area += length_vector(cross_vectors(v1, v3))
                proj_area = area * 0.25
                proj_area_total += proj_area
                # this should become a method
                vol_i = proj_area*(height - zi)
                volume += vol_i

        self.fill = True
        self.fill_volume = volume
        print('Proj area total of shape', proj_area_total)
        self.fill_ro = fill_ro

        return

    def compute_fill_weight(self):
        """Compute and returns the volume of fill in the structure.

        Returns
        -------
        volume : float
            The volume.
        """
        return self.fill_volume * self.fill_ro



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas_tno.rhino.diagramartist import DiagramArtist
from compas.colors import Color
from compas.utilities import is_color_rgb


__all__ = ['ForceArtist']


class ForceArtist(DiagramArtist):
    """Artist for force diagrams in AGS.

    Parameters
    ----------
    form: compas_tno.diagrams.FormDiagram
        The form diagram to draw.

    """

    def __init__(self, force, layer=None):
        super(ForceArtist, self).__init__(force, layer=layer)

        self._mesh = None
        self._vertices = None
        self._edges = None
        self._faces = None
        self._color = None
        self._vertex_xyz = None
        self._vertex_color = None
        self._vertex_text = None
        self._vertex_size = None
        self._edge_color = None
        self._edge_text = None
        self._edge_width = None
        self._face_color = None
        self._face_text = None

        self._vertexcollection = None
        self._edgecollection = None
        self._facecollection = None
        self._vertexnormalcollection = None
        self._facenormalcollection = None
        self._vertexlabelcollection = None
        self._edgelabelcollection = None
        self._facelabelcollection = None
        self.default_color = Color.black()

        self.mesh = force

    @property
    def mesh(self):
        return self._mesh

    @mesh.setter
    def mesh(self, mesh):
        self._mesh = mesh
        self._vertex_xyz = None

    @property
    def vertices(self):
        if self._vertices is None:
            self._vertices = list(self.mesh.vertices())
        return self._vertices

    @vertices.setter
    def vertices(self, vertices):
        self._vertices = vertices

    @property
    def edges(self):
        if self._edges is None:
            self._edges = list(self.mesh.edges())
        return self._edges

    @edges.setter
    def edges(self, edges):
        self._edges = edges

    @property
    def faces(self):
        if self._faces is None:
            self._faces = list(self.mesh.faces())
        return self._faces

    @faces.setter
    def faces(self, faces):
        self._faces = faces

    @property
    def color(self):
        if not self._color:
            self._color = self.default_color
        return self._color

    @color.setter
    def color(self, color):
        if is_color_rgb(color):
            self._color = color

    @property
    def vertex_xyz(self):
        if self._vertex_xyz is None:
            return {vertex: self.mesh.vertex_attributes(vertex, 'xyz') for vertex in self.mesh.vertices()}
        return self._vertex_xyz

    @vertex_xyz.setter
    def vertex_xyz(self, vertex_xyz):
        self._vertex_xyz = vertex_xyz

    @property
    def vertex_color(self):
        if self._vertex_color is None:
            self._vertex_color = {vertex: self.default_vertexcolor for vertex in self.mesh.vertices()}
        return self._vertex_color

    @vertex_color.setter
    def vertex_color(self, vertex_color):
        if isinstance(vertex_color, dict):
            self._vertex_color = vertex_color
        elif is_color_rgb(vertex_color):
            self._vertex_color = {vertex: vertex_color for vertex in self.mesh.vertices()}

    @property
    def vertex_text(self):
        if self._vertex_text is None:
            self._vertex_text = {vertex: str(vertex) for vertex in self.mesh.vertices()}
        return self._vertex_text

    @vertex_text.setter
    def vertex_text(self, text):
        if text == 'key':
            self._vertex_text = {vertex: str(vertex) for vertex in self.mesh.vertices()}
        elif text == 'index':
            self._vertex_text = {vertex: str(index) for index, vertex in enumerate(self.mesh.vertices())}
        elif isinstance(text, dict):
            self._vertex_text = text

    @property
    def vertex_size(self):
        if not self._vertex_size:
            self._vertex_size = {vertex: self.default_vertexsize for vertex in self.mesh.vertices()}
        return self._vertex_size

    @vertex_size.setter
    def vertex_size(self, vertexsize):
        if isinstance(vertexsize, dict):
            self._vertex_size = vertexsize
        elif isinstance(vertexsize, (int, float)):
            self._vertex_size = {vertex: vertexsize for vertex in self.mesh.vertices()}

    @property
    def edge_color(self):
        if self._edge_color is None:
            self._edge_color = {edge: self.default_edgecolor for edge in self.mesh.edges()}
        return self._edge_color

    @edge_color.setter
    def edge_color(self, edge_color):
        if isinstance(edge_color, dict):
            self._edge_color = edge_color
        elif is_color_rgb(edge_color):
            self._edge_color = {edge: edge_color for edge in self.mesh.edges()}

    @property
    def edge_text(self):
        if self._edge_text is None:
            self._edge_text = {edge: "{}-{}".format(*edge) for edge in self.mesh.edges()}
        return self._edge_text

    @edge_text.setter
    def edge_text(self, text):
        if text == 'key':
            self._edge_text = {edge: "{}-{}".format(*edge) for edge in self.mesh.edges()}
        elif text == 'index':
            self._edge_text = {edge: str(index) for index, edge in enumerate(self.mesh.edges())}
        elif isinstance(text, dict):
            self._edge_text = text

    @property
    def edge_width(self):
        if not self._edge_width:
            self._edge_width = {edge: self.default_edgewidth for edge in self.mesh.edges()}
        return self._edge_width

    @edge_width.setter
    def edge_width(self, edgewidth):
        if isinstance(edgewidth, dict):
            self._edge_width = edgewidth
        elif isinstance(edgewidth, (int, float)):
            self._edge_width = {edge: edgewidth for edge in self.mesh.edges()}

    @property
    def face_color(self):
        if self._face_color is None:
            self._face_color = {face: self.default_facecolor for face in self.mesh.faces()}
        return self._face_color

    @face_color.setter
    def face_color(self, face_color):
        if isinstance(face_color, dict):
            self._face_color = face_color
        elif is_color_rgb(face_color):
            self._face_color = {face: face_color for face in self.mesh.faces()}

    @property
    def face_text(self):
        if self._face_text is None:
            self._face_text = {face: str(face) for face in self.mesh.faces()}
        return self._face_text

    @face_text.setter
    def face_text(self, text):
        if text == 'key':
            self._face_text = {face: str(face) for face in self.mesh.faces()}
        elif text == 'index':
            self._face_text = {face: str(index) for index, face in enumerate(self.mesh.faces())}
        elif isinstance(text, dict):
            self._face_text = text



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc
import compas_rhino


__commandname__ = "TNO_scene_redraw"


def RunCommand(is_interactive):

    if 'TNO' not in sc.sticky:
        compas_rhino.display_message('TNO has not been initialised yet.')
        return

    scene = sc.sticky['TNO']['scene']
    if not scene:
        return

    scene.update()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    RunCommand(True)



from compas.datastructures import Mesh
from compas.datastructures import mesh_bounding_box_xy

from compas_tno.diagrams.diagram_arch import create_arch_form_diagram
from compas_tno.diagrams.diagram_arch import create_linear_form_diagram
from compas_tno.diagrams.diagram_rectangular import create_cross_form
from compas_tno.diagrams.diagram_rectangular import create_cross_diagonal
from compas_tno.diagrams.diagram_rectangular import create_cross_with_diagonal
from compas_tno.diagrams.diagram_rectangular import create_fan_form
from compas_tno.diagrams.diagram_circular import create_circular_radial_form
from compas_tno.diagrams.diagram_circular import create_circular_radial_spaced_form
from compas_tno.diagrams.diagram_circular import create_circular_spiral_form
from compas_tno.diagrams.diagram_rectangular import create_ortho_form
from compas_tno.diagrams.diagram_rectangular import create_parametric_form
from compas_tno.diagrams.diagram_rectangular import create_delta_form

from compas_tna.diagrams import FormDiagram

from compas.geometry import Point

import math


class FormDiagram(FormDiagram):
    """The ``FormDiagram`` represents the projection of the internal force network within the structure.

    The class imports the attributes and set ups from ``compas_tna.diagrams.FormDiagram`` and include some
    functionalities useful for the assessment of masonry structures. Most relevant edge and vertex attributes
    are listed below:

    Attributes
    ----------
    *   Vertex attributes

    x : float
        x-coordinate of the vertex.
    y : float
        y-coordinate of the vertex.
    z : float
        z-coordinate of the vertex.
    is_roller : bool
        Whether or not the attibute is set to be a roller. Takes only horizontal reaction.
    tubmax : float
        Maximum allowable increase in the extrados
    tlbmax : float
        Maximum allowable increase in the intradoss
    tub_reac : float
        Maximum allowable increase in the horizontal projection of the reaction.

    *   Edge attributes

    q : float
        Force density of an edge.
    f : float
        Force magnitude of an edge.
    is_ind : bool
        Whether or not the edge is an independent edge.
    qmin : float
        The minimum force density magnitude.
    qmax : float
        The maximum force density magnitude.

    Examples
    --------
    >>> from compas_tno.diagrams import FormDiagram
    >>> data = {'type': 'fan_fd', 'xy_span': [[0, 10], [0, 10]], 'discretisation': [10, 10], 'fix': 'corners'}
    >>> form = FormDiagram.from_library(data)

    """

    def __init__(self):
        super(FormDiagram, self).__init__()
        self.update_default_vertex_attributes({
            'is_roller': False,
            'tubmax': None,
            'tlbmax': None,
            'tub_reac': None,
        })
        self.update_default_edge_attributes({
            'q': -1.0,
            'f': -1.0,
            'is_symmetry': False,
            'is_ind': False,
            'qmin': -1e+4,
            'qmax': 1e-8
        })
        self.args = None
        self.attributes['loadpath'] = 0.0
        self.attributes['indset'] = None
        self.parameters = {
            'type': None,
            'discretisation': None,
            'E': None,
            'Ah': None,
            # 'x0': None,
            # 'discretisation': 20,
            # 'xy_span': [[0,10],[0,10]],
            # 'discretisation': [10,10],
            # 'fix': 'corners',
            # 'r_oculus:' None,
        }

    # --------------------------------------------------------------- #
    # -----------------------CONSTRUCTORS---------------------------- #
    # --------------------------------------------------------------- #

    @classmethod
    def from_library(cls, data):
        """ Construct a FormDiagram the library of FormDiagrams available.

        Parameters
        ----------
        data : dict
            Dictionary with information to generate the FormDiagram.

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        # General parameters
        form_type = data['type']
        discretisation = data.get('discretisation', 10)

        # Arch parameters
        H = data.get('H', 1.0)
        L = data.get('L', 2.0)
        x0 = data.get('x0', 0.0)

        # Rectangular vault parameters
        xy_span = data.get('xy_span', [[0.0, 10.0], [0.0, 10.0]])
        fix = data.get('fix', 'corners')

        # Circular parameters
        center = data.get('center', [5.0, 5.0])
        radius = data.get('radius', 5.0)
        r_oculus = data.get('r_oculus', 0.0)
        diagonal = data.get('diagonal')
        partial_diagonal = data.get('partial_diagonal')
        partial_bracing_modules = data.get('partial_bracing_modules')
        fix = data.get('fix', 'corners')
        lambd = data.get('lambd', 0.5)
        delta = data.get('delta', 0.5)

        if form_type == 'arch':
            form = create_arch_form_diagram(cls(), H=H, L=L, x0=x0, discretisation=discretisation)
        elif form_type == 'linear_arch' or form_type == 'pointed_arch':
            form = create_linear_form_diagram(cls(), L=L, x0=x0, discretisation=discretisation)
        elif form_type == 'cross_fd':
            form = create_cross_form(cls(), xy_span=xy_span, discretisation=discretisation, fix=fix)
        elif form_type == 'cross_diagonal':
            form = create_cross_diagonal(cls(), xy_span=data['xy_span'], discretisation=discretisation,
                                         partial_bracing_modules=partial_bracing_modules, fix=fix)
        elif form_type == 'cross_with_diagonal':
            form = create_cross_with_diagonal(cls(), xy_span=data['xy_span'], discretisation=discretisation, fix=fix)
        elif form_type == 'fan_fd':
            form = create_fan_form(cls(), xy_span=data['xy_span'], discretisation=discretisation, fix=fix)
        elif form_type == 'ortho' or form_type == 'ortho_fd':
            form = create_ortho_form(cls(), xy_span=xy_span, discretisation=discretisation, fix=fix)
        elif form_type == 'radial_fd':
            form = create_circular_radial_form(cls(), center=center, radius=radius, discretisation=discretisation,
                                               r_oculus=r_oculus, diagonal=diagonal, partial_diagonal=partial_diagonal)
        elif form_type == 'radial_spaced_fd':
            form = create_circular_radial_spaced_form(cls(), center=center, radius=radius, discretisation=discretisation,
                                                      r_oculus=r_oculus, diagonal=diagonal, partial_diagonal=partial_diagonal)
        elif form_type == 'spiral_fd':
            form = create_circular_spiral_form(cls(), center=center, radius=radius, discretisation=discretisation, r_oculus=r_oculus)
        elif form_type == 'parametric_form':
            form = create_parametric_form(cls(), xy_span=xy_span, discretisation=discretisation, lambd=lambd, fix=fix)
        elif form_type == 'delta_form':
            form = create_delta_form(cls(), xy_span=xy_span, discretisation=discretisation, delta=delta, fix=fix)
        else:
            raise NotImplementedError()

        form.parameters = data

        return form

    @classmethod
    def create_arch(cls, H=1.00, L=2.00, x0=0.0, discretisation=100):
        """Construct a FormDiagram based on an arch linear discretisation.
        Note: The nodes of the form diagram are spaced following a projection in a semicircular arch.

        Parameters
        ----------
        H : float, optional
            Height of the arch, by default 1.00
        L : float, optional
            Span of the arch, by default 2.00
        x0 : float, optional
            Initial coordiante of the arch, by default 0.0
        discretisation : int, optional
            Numbers of nodes to be considered in the form diagram, by default 100

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        data = {'type': 'arch', 'L': L, 'H': H, 'x0': x0, 'discretisation': discretisation}

        return cls().from_library(data)

    @classmethod
    def create_linear_form_diagram(cls, L=2.0, x0=0.0, discretisation=100):
        """ Helper to create a arch linear form-diagram with equaly spaced (in 2D) nodes.

        Parameters
        ----------
        L : float, optional
            Span of the arch, by default 2.00
        x0 : float, optional
            Initial coordiante of the arch, by default 0.0
        discretisation : int, optional
            Numbers of nodes to be considered in the form diagram, by default 100

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            FormDiagram generated according to the parameters.
        """

        data = {'type': 'linear_arch', 'L': L, 'x0': x0, 'discretisation': discretisation}

        return cls().from_library(data)

    @classmethod
    def create_cross_form(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=10, fix='corners'):
        """Construct a FormDiagram based on cross discretiastion with orthogonal arrangement and diagonal.

        Parameters
        ----------
        xy_span : list, optional
            List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
        discretisation : int, optional
            Set the density of the grid in x and y directions, by default 10
        fix : str, optional
            Option to select the constrained nodes: 'corners', 'all' are accepted., by default 'corners'

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        data = {'type': 'cross_fd', 'xy_span': xy_span, 'discretisation': discretisation, 'fix': fix}

        return cls().from_library(data)

    @classmethod
    def create_cross_diagonal(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=10, partial_bracing_modules=None, fix='corners'):
        """Construct a FormDiagram based on a mixture of cross and fan discretiastion

        Parameters
        ----------
        xy_span : list, optional
            List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
        discretisation : int, optional
            Set the density of the grid in x and y directions, by default 10
        partial_bracing_modules : str, optional
            If partial bracing modules are included, by default None
        fix : str, optional
            Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        data = {'type': 'cross_diagonal', 'xy_span': xy_span, 'discretisation': discretisation, 'partial_bracing_modules': partial_bracing_modules, 'fix': fix}

        return cls().from_library(data)

    @classmethod
    def create_cross_with_diagonal(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=10, fix='corners'):
        """ Construct a FormDiagram based on cross discretiastion with diagonals.

        Parameters
        ----------
        xy_span : list, optional
            List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
        discretisation : int, optional
            Set the density of the grid in x and y directions, by default 10
        fix : str, optional
            Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        data = {'type': 'cross_with_diagonal', 'xy_span': xy_span, 'discretisation': discretisation, 'fix': fix}

        return cls().from_library(data)

    @classmethod
    def create_fan_form(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=[10, 10], fix='corners'):
        """ Helper to construct a FormDiagram based on fan discretiastion with straight lines to the corners.

        Parameters
        ----------
        xy_span : list, optional
            List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
        discretisation : int, optional
            Set the density of the grid in x and y directions, by default 10
        fix : str, optional
            Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        data = {'type': 'fan_fd', 'xy_span': xy_span, 'discretisation': discretisation, 'fix': fix}

        return cls().from_library(data)

    @classmethod
    def create_parametric_form(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=10, lambd=0.5, fix='corners'):
        """ Helper to construct a FormDiagram envelop that includes fan and cross diagrams. The diagram is modified by the parameter lambda.

        Parameters
        ----------
        xy_span : [[float, float], [float, float]], optional
            List with initial- and end-points of the vault, by default, by default [[0.0, 10.0], [0.0, 10.0]]
        discretisation : int, optional
            Set the density of the grid in x and y directions, by default 10
        lambd : float, optional
            Inclination of the arches in the diagram (0.0 will result in cross and 1.0 in fan diagrams), by default 0.5
        fix : str, optional
            Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        data = {'type': 'parametric_form', 'xy_span': xy_span, 'discretisation': discretisation, 'fix': fix, 'lambd': lambd}

        return cls().from_library(data)

    @classmethod
    def create_delta_form(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=10, delta=0.5, fix='corners'):
        """ Helper to construct a FormDiagram based on fan discretiastion with straight lines to the corners.

        Parameters
        ----------
        xy_span : [[float, float], [float, float]], optional
            List with initial- and end-points of the vault, by default, by default [[0.0, 10.0], [0.0, 10.0]]
        discretisation : int, optional
            Set the density of the grid in x and y directions, by default 10
        lambd : float, optional
            Inclination of the arches in the diagram (0.0 will result in cross and 1.0 in fan diagrams), by default 0.5
        fix : str, optional
            Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        data = {'type': 'delta_form', 'xy_span': xy_span, 'discretisation': discretisation, 'fix': fix, 'delta': delta}

        return cls().from_library(data)

    @classmethod
    def create_ortho_form(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=[10, 10], fix='corners'):
        """ Helper to construct a FormDiagram based on a simple orthogonal discretisation.

        Parameters
        ----------
        xy_span : list, optional
            List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
        discretisation : int, optional
            Set the density of the grid in x and y directions, by default 10
        fix : str, optional
            Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        data = {'type': 'ortho_fd', 'xy_span': xy_span, 'discretisation': discretisation, 'fix': fix}

        return cls().from_library(data)

    @classmethod
    def create_circular_radial_form(cls, center=[5.0, 5.0], radius=5.0, discretisation=[8, 20], r_oculus=0.0, diagonal=False, partial_diagonal=False):
        """Construct a circular radial FormDiagram with hoops not equally spaced in plan.

        Parameters
        ----------
        center : list, optional
            Planar coordinates of the form-diagram [xc, yc], by default [5.0, 5.0]
        radius : float, optional
            Radius of the form diagram, by default 5.0
        discretisation : list, optional
            Number of hoops, and of parallels of the dome form diagram], by default [8, 20]
        r_oculus : float, optional
            Value of the radius of the oculus, if no oculus is present should be set to zero, by default 0.0
        diagonal : bool, optional
            Activate diagonal in the quads, by default False
        partial_diagonal : bool, optional
            Activate partial diagonal in the quads, by default False

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        data = {'type': 'radial_fd', 'center': center, 'radius': radius, 'discretisation': discretisation,
                'r_oculus': r_oculus, 'diagonal': diagonal, 'partial_diagonal': partial_diagonal}

        return cls().from_library(data)

    @classmethod
    def create_circular_radial_spaced_form(cls, center=[5.0, 5.0], radius=5.0, discretisation=[8, 20], r_oculus=0.0, diagonal=False, partial_diagonal=False):
        """ Construct a circular radial FormDiagram with hoops not equally spaced in plan, but equally spaced with regards to the projection on a hemisphere.

        Parameters
        ----------
        center : list, optional
            Planar coordinates of the form-diagram [xc, yc], by default [5.0, 5.0]
        radius : float, optional
            Radius of the form diagram, by default 5.0
        discretisation : list, optional
            Number of hoops, and of parallels of the dome form diagram], by default [8, 20]
        r_oculus : float, optional
            Value of the radius of the oculus, if no oculus is present should be set to zero, by default 0.0
        diagonal : bool, optional
            Activate diagonal in the quads, by default False
        partial_diagonal : bool, optional
            Activate partial diagonal in the quads, by default False

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        data = {'type': 'radial_spaced_fd', 'center': center, 'radius': radius, 'discretisation': discretisation,
                'r_oculus': r_oculus, 'diagonal': diagonal, 'partial_diagonal': partial_diagonal}

        return cls().from_library(data)

    @classmethod
    def create_circular_spiral_form(cls, center=[5.0, 5.0], radius=5.0, discretisation=[8, 20], r_oculus=0.0):
        """ Construct a circular radial FormDiagram with hoops not equally spaced in plan, but equally spaced with regards to the projection on a hemisphere.

        Parameters
        ----------
        center : list, optional
            Planar coordinates of the form-diagram [xc, yc], by default [5.0, 5.0]
        radius : float, optional
            Radius of the form diagram, by default 5.0
        discretisation : list, optional
            Number of hoops, and of parallels of the dome form diagram], by default [8, 20]
        r_oculus : float, optional
            Value of the radius of the oculus, if no oculus is present should be set to zero, by default 0.0

        Returns
        -------
        FormDiagram
            The FormDiagram created.
        """

        data = {'type': 'spiral_fd', 'center': center, 'radius': radius, 'discretisation': discretisation, 'r_oculus': r_oculus}

        return cls().from_library(data)

    @classmethod
    def from_assembly(self):
        NotImplementedError

    @classmethod
    def from_singular(cls, json_path):
        """Create form diagram from singular dense mesh.

        Parameters
        ----------
        json_path : str
            Path of the mesh saved in JSON format.

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        from compas.datastructures import mesh_delete_duplicate_vertices
        mesh = Mesh.from_json(json_path)
        mesh_delete_duplicate_vertices(mesh)
        form = cls().from_mesh(mesh)
        form.set_boundary_supports()
        form.delete_boundary_edges()

        return form

    @classmethod
    def from_triangle(cls, boundary_points, area=0.5, angle=30):
        """Generates a Form Diagram from triangle

        Parameters
        ----------
        boundary_points : [list]
            Boundary point of pattern
        area : float, optional
            Average area of the trangles, by default 0.5
        angle : int, optional
            minimum angle in the triangless, by default 30

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        from compas_triangle.delaunay import conforming_delaunay_triangulation
        vertices, faces = conforming_delaunay_triangulation(boundary_points + boundary_points[:1], angle=angle, area=area)
        mesh = Mesh.from_vertices_and_faces(vertices, faces)
        form = cls().from_mesh(mesh)

        return form

    @classmethod
    def from_skeleton(cls, data):
        """Create form diagram from compas_skeleton

        Parameters
        ----------
        data : dict
            Dictionary with the data for crerating the skeleton mesh.

        Returns
        -------
        :class:`~compas_tno.diagrams.FormDiagram`
            The FormDiagram created.
        """

        from compas_skeleton.datastructure import Skeleton
        from compas.geometry import matrix_from_scale_factors
        from compas.datastructures import mesh_transform

        skeleton_type = data['type']

        if skeleton_type == 'circular':

            point = data.get('center', [0, 0, 0])
            subdivisions = data.get('subdivision', 3)
            radius = data.get('radius', 5.0)

            skeleton = Skeleton.from_center_point(point)
            skeleton.node_width = 7.5
            skeleton.update_mesh_vertices_pos()
            skeleton.vertex_attribute(0, 'z', 10.0)
            skeleton.subdivide(subdivisions)
            skeleton_mesh = skeleton.to_mesh()

            max_x = 0
            for key in skeleton_mesh.vertices():
                x, _, _ = skeleton_mesh.vertex_coordinates(key)
                if x > max_x:
                    max_x = x
            scale = radius / max_x
            T = matrix_from_scale_factors([scale, scale, 0.0])
            mesh_transform(skeleton_mesh, T)
            form = cls().from_mesh(skeleton_mesh)
            form.set_boundary_supports()
            form.delete_boundary_edges()

        return form

    # --------------------------------------------------------------------------
    # Plot
    # --------------------------------------------------------------------------

    def plot(self):
        """Plot a form diagram with a plotter with all the default settings."""
        from compas_plotters import Plotter
        plotter = Plotter(figsize=(8, 8))
        plotter.add(self)
        plotter.show()

    # --------------------------------------------------------------- #
    # -----------------------ULTILITIES------------------------------ #
    # --------------------------------------------------------------- #

    def q(self):
        """Return the force densities ``q`` of the diagram."""
        return [self.edge_attribute(edge, 'q') for edge in self.edges_where({'_is_edge': True})]

    def xy(self):
        """Return the ``xy`` coordinates of the diagram."""
        return self.vertices_attributes('xy')

    def fixed(self):
        """Return the keys of the fixed vertices of the diagram."""
        return list(self.vertices_where({'is_fixed': True}))

    def fixed_x(self):
        """Return the keys of the fixed-in-x vertices of the diagram."""
        return list(self.vertices_where({'is_fixed_x': True, 'is_fixed': False}))

    def fixed_y(self):
        """Return the keys of the fixed-in-y vertices of the diagram."""
        return list(self.vertices_where({'is_fixed_y': True, 'is_fixed': False}))

    def ind(self):
        """Return the identifiers for the edges selected as independents on the diagram."""
        return list(self.edges_where({'is_ind': True}))

    def number_of_real_edges(self):
        """Return the identifiers for the edges selected as independents on the diagram."""
        return len(list(self.edges_where({'_is_edge': True})))

    def update_f(self):
        """Update attributes force ``f`` based on force densities ``q``and lengths ``l``."""

        for u, v in self.edges():
            f = self.edge_attribute((u, v), 'q') * self.edge_length(u, v)
            self.edge_attribute((u, v), 'f', f)

    def add_pz(self, key, force=0.0):
        """_summary_

        Parameters
        ----------
        key : int
            Edge identifier
        force : float, optional
            Load to add at the node in the attribute 'pz', by default 0.0

        """

        pz0 = self.vertex_attribute(key, 'pz')
        self.vertex_attribute(key, 'pz', pz0 + force)

    def assign_inds(self, edges=None, printout=True):
        """Assign manually independent edges to the pattern

        Parameters
        ----------
        edges : [edges], optional
            List of edges to be assigned as independents, by default None
        printout : bool, optional
            Whether the number of assigned independent edges should appear, by default True

        """

        if not edges:
            return

        self.edges_attribute('is_ind', False)

        i = 0
        for edge in edges:
            u, v = edge
            try:
                self.edge_attribute((u, v), 'is_ind', True)
            except BaseException:
                self.edge_attribute((v, u), 'is_ind', True)
            i += 1

        self.store_indset(edges)

        if printout:
            print('Assigned {} independent edges to the pattern.'.format(i))

    def store_indset(self, edges=None):
        """Store the indset with the information about the independent edges (used within algorithms)

        Parameters
        ----------
        edges : [edges], optional
            List of edges to be assigned as independents, by default None

        """

        if not edges:
            self.attributes['indset'] = None
            return

        points = []
        for edge in edges:
            points.append(Point(*(self.edge_midpoint(*edge)[:2] + [0])))

        self.attributes['indset'] = points

    def update_indset(self):
        """Update the indset with the information about the independent edges (used within algorithms)
        """

        edges = list(self.edges_where({'is_ind': True}))

        if not edges:
            self.attributes['indset'] = None
            return

        self.store_indset(edges)

    def tributary_dict(self):
        """Make a tributary dictionary to help in the calculation of the tributary weights to the nodes afterwards.

        Returns
        -------
        tributary_dict
            The tributary area dictionary
        """

        tributary_dict = {}
        vertex_index = self.vertex_index()

        for key in self.vertices():
            i = vertex_index[key]
            tributary_dict[i] = {}
            for nbr in self.halfedge[key]:
                j = vertex_index[nbr]
                tributary_dict[i][j] = []
                fkey = self.halfedge[key][nbr]
                if fkey is not None:
                    facekeys = self.face_vertices(fkey)
                    faceindices = [vertex_index[v] for v in facekeys]
                    tributary_dict[i][j].append(faceindices)

                fkey = self.halfedge[nbr][key]
                if fkey is not None:
                    facekeys = self.face_vertices(fkey)
                    faceindices = [vertex_index[v] for v in facekeys]
                    tributary_dict[i][j].append(faceindices)

        return tributary_dict

    def tributary_matrices(self, sparse=False):
        """Returns the Matrices used for computation of the tributary area.

        Parameters
        ----------
        sparse : bool, optional
            If the matrices should be returned as sparse matrices (see ``scipy.linalg.sparse``), by default False

        Returns
        -------
        F : array [f x n]
            Linear transformation from ``X`` [n x 3] to ``c`` [f x 3] with the position of the centroids
        V0 : array [g x n]
            Mark the influence of the original point in the calculation
        V1 : array [g x n]
            Mark the influence of the neighbor points the calculation
        V2 : array [g x f]
            Mark the influence of the centroid points the calculation
        """

        from numpy import zeros
        from numpy import vstack

        n = self.number_of_vertices()
        f = self.number_of_faces()

        vertex_index = self.vertex_index()
        face_index = {}

        F = zeros((f, n))
        V0 = zeros((0, n))
        V1 = zeros((0, n))
        V2 = zeros((0, f))

        for i, fkey in enumerate(self.faces()):
            face_index[fkey] = i
            faceindices = [vertex_index[v] for v in self.face_vertices(fkey)]
            np = float(len(faceindices))
            for j in faceindices:
                F[i, j] = 1.0/np

        ig = 0
        for key in self.vertices():
            i = vertex_index[key]
            v0i = zeros((1, n))
            v0i[0, i] = 1.0

            for nbr in self.halfedge[key]:
                j = vertex_index[nbr]

                fkey = self.halfedge[key][nbr]
                if fkey is not None:
                    findex = face_index[fkey]
                    facekeys = self.face_vertices(fkey)
                    faceindices = [vertex_index[v] for v in facekeys]

                    v1i = zeros((1, n))
                    v1i[0, j] = 1.0
                    V1 = vstack([V1, v1i])

                    v2i = zeros((1, f))
                    v2i[0, findex] = 1.0
                    V2 = vstack([V2, v2i])

                    V0 = vstack([V0, v0i])

                    ig += 1

                fkey = self.halfedge[nbr][key]
                if fkey is not None:
                    findex = face_index[fkey]
                    facekeys = self.face_vertices(fkey)
                    faceindices = [vertex_index[v] for v in facekeys]

                    v1i = zeros((1, n))
                    v1i[0, j] = 1.0
                    V1 = vstack([V1, v1i])

                    v2i = zeros((1, f))
                    v2i[0, findex] = 1.0
                    V2 = vstack([V2, v2i])

                    V0 = vstack([V0, v0i])

                    ig += 1

        return F, V0, V1, V2

    def update_lumped_weights(self, thickness=0.5, density=20.0):
        """Update the lumped weights based on the curent geometry of the thrust network
        The loads are computed based on the tributary area times the thickness times the density.
        For variable thickness the nodal attribute `thk` is considered.

        Parameters
        ----------
        thickness : float, optional
            The thickness of the problem, by default 0.50
            If None is passed, the thickness is taken from the nodal attribute `thk`
        density : float, optional
            The density assumed, by default 20.0

        Returns
        -------
        None
            Loads are updated in place.
        """

        from compas_tno.utilities import apply_selfweight_from_thrust
        apply_selfweight_from_thrust(self, thickness=thickness, density=density)

    def overview_forces(self):
        """Print an overview of the forces in the ``Form Diagram``."""

        self.update_f()

        f = []
        q = []
        z = []
        pz = 0

        lp = 0

        for u, v in self.edges_where({'_is_edge': True}):
            qi = self.edge_attribute((u, v), 'q')
            li = self.edge_length(u, v)
            lp += qi*li**2
            q.append(qi)
            f.append(qi*li)

        print('='*20)
        print('Overview on forces:')

        print('q: {0:.3f} : {1:.3f}'.format(float(min(q)), float(max(q))))
        print('f: {0:.3f} : {1:.3f}'.format(float(min(f)), float(max(f))))
        for key in self.vertices():
            z.append(self.vertex_attribute(key, 'z'))
            pz += self.vertex_attribute(key, 'pz')
        print('z: {0:.3f} : {1:.3f}'.format(float(min(z)), float(max(z))))
        print('pz: {0:.3f}'.format(pz))
        print('lp: {0:.3f}'.format(lp))
        self.attributes['loadpath'] = lp

        return

    def thrust(self):
        """Computes the the (horizontal) thrust"""

        thrust = 0
        for key in self.vertices_where({'is_fixed': True}):
            rx = self.vertex_attribute(key, '_rx')
            ry = self.vertex_attribute(key, '_ry')
            r = math.sqrt(rx**2 + ry**2)
            thrust += r

        return thrust

    def lumped_swt(self):
        """Returns the lumped selfweight on the nodes."""

        swt = 0
        for key in self.vertices():
            pz = self.vertex_attribute(key, 'pz')
            swt += pz

        return swt

    def distance_target(self):
        """ Compute the squared vertical distance from a target."""

        f = 0
        for key in self.vertices():
            z = self.vertex_attribute(key, 'z')
            s = self.vertex_attribute(key, 'target')
            f += (z - s)**2

        return f

    def loadpath(self):
        """ Compute loadpath in the current configuration based on the force attributes stored in the ``FormDiagram``. """

        lp = 0
        for u, v in self.edges_where({'_is_edge': True}):
            qi = abs(self.edge_attribute((u, v), 'q'))
            li = self.edge_length(u, v)
            lp += qi*li**2

        self.attributes['loadpath'] = lp

        return lp

    def number_of_independents(self):
        """ Compute the number of independent edges."""
        return len(list(self.edges_where({'is_ind': True})))

    # --------------------------------------------------------------- #
    # -----------------------BOUNDARIES------------------------------ #
    # --------------------------------------------------------------- #

    def number_of_supports(self):
        """ Compute the number of supports."""
        return len(list(self.vertices_where({'is_fixed': True})))

    def delete_boundary_edges(self, delete_corner_vertex=True):
        """
        Delete boundary edges on a diagram.

        Parameters
        ----------
        delete_corner_vertex : bool, optional
            Delete the corner vertices. The default value is ``True``.

        Returns
        -------
        None
            The ``FormDiagram`` is modified in place.

        """

        for u, v in self.edges_on_boundary():
            if self.vertex_attribute(u, 'is_fixed') and self.vertex_attribute(v, 'is_fixed'):
                self.edge_attribute((u, v), '_is_edge', False)

        if delete_corner_vertex:
            corners = [key for key in self.corners()]
            for key in corners:
                neighbors = self.vertex_neighbors(key)
                print(neighbors)
                n1 = self.vertex_neighbors(neighbors[0])
                n2 = self.vertex_neighbors(neighbors[1])
                intersection_vertex = list(set.intersection(set(n1), set(n2)))
                sum_vertices = neighbors + intersection_vertex
                new_face = []
                for key_ in sum_vertices:
                    if key_ != key:
                        new_face.append(key_)
                self.delete_vertex(key)
                self.add_face(new_face)
                self.edge_attribute((neighbors[0], neighbors[1]), '_is_edge', False)

        return self

    def set_boundary_supports(self):
        """ Set all node on the boundary of a pattern as supports."""

        for key in self.vertices_on_boundary():
            self.vertex_attribute(key, 'is_fixed', True)

        return

    def set_boundary_rollers(self, max_rx=[0.0, 0.0], max_ry=[0.0, 0.0], total_rx=None, total_ry=None):
        """
        Set boundary vertices as rollers.

        Parameters
        ----------
        max_rx : list, optional
            Assign the maximum reaction in x to the vertices in the left and right boundary. The default value is ``[0.0, 0.0]``.
        max_ry : list, optional
            Assign the maximum reaction in y to the vertices in the top and bottom boundary. The default value is ``[0.0, 0.0]``.
        total_rx : list, optional
            Assign the maximum total reaction in x of the vertices in the right or left boundary. The default value is ``None``, in which the ``max_rx`` is used for all vertices.
        total_ry : list, optional
            Assign the maximum total reaction in y of the vertices in the top or bottom boundary.The default value is ``None``, in which the ``max_ry`` is used for all vertices.

        Returns
        -------
        None
            The ``FormDiagram`` is modified in place.

        """
        corners = mesh_bounding_box_xy(self)
        xs = [point[0] for point in corners]
        ys = [point[1] for point in corners]
        xlimits = [min(xs), max(xs)]
        ylimits = [min(ys), max(ys)]
        if not total_rx:
            nx0 = 0
            nx1 = 0
            for key in self.vertices_where({'x': xlimits[0]}):
                nx0 += 1
            for key in self.vertices_where({'x': xlimits[1]}):
                nx1 += 1
            max_rx = [total_rx[0]/nx0, total_rx[1]/nx1]
        if not total_ry:
            ny0 = 0
            ny1 = 0
            for key in self.vertices_where({'y': ylimits[0]}):
                ny0 += 1
            for key in self.vertices_where({'y': ylimits[1]}):
                ny1 += 1
            max_ry = [total_ry[0]/ny0, total_ry[1]/ny1]
        for key in self.vertices_on_boundary():
            if self.vertex_attribute(key, 'is_fixed') is False:
                x, y, _ = self.vertex_coordinates(key)
                if x == xlimits[0]:
                    self.vertex_attribute(key, 'rol_x', True)
                    self.vertex_attribute(key, 'max_rx', max_rx[0])
                elif x == xlimits[1]:
                    self.vertex_attribute(key, 'rol_x', True)
                    self.vertex_attribute(key, 'max_rx', max_rx[1])
                elif y == ylimits[0]:
                    self.vertex_attribute(key, 'rol_y', True)
                    self.vertex_attribute(key, 'max_ry', max_ry[0])
                elif y == ylimits[1]:
                    self.vertex_attribute(key, 'rol_y', True)
                    self.vertex_attribute(key, 'max_ry', max_ry[1])

        return

    # --------------------------------------------------------------- #
    # --------------------------SYMMETRY----------------------------- #
    # --------------------------------------------------------------- #

    def number_of_sym_edges(self, printout=False):
        """ Compute the symmetric edges."""

        i_sym_max = 0
        for u, v in self.edges_where({'_is_edge': True}):
            try:
                i_sym = self.edge_attribute((u, v), 'sym_key')
            except BaseException:
                print('Warning, no symmetry relation found!')
            if i_sym > i_sym_max:
                i_sym_max = i_sym
        i_sym_max += 1

        return i_sym_max

    def number_of_sym_vertices(self):
        """ Compute the symmetric vertices."""

        i_sym_max = 0
        for key in self.vertices():
            try:
                i_sym = self.vertex_attribute(key, 'sym_key')
            except BaseException:
                print('Warning, no symmetry relation found!')
            if i_sym > i_sym_max:
                i_sym_max = i_sym
        i_sym_max += 1

        return i_sym_max

    def number_of_sym_supports(self):
        """ Compute the symmetric supports."""

        i_sym_max = 0
        for key in self.vertices_where({'is_fixed': True}):
            try:
                i_sym = self.vertex_attribute(key, 'sym_key')
            except BaseException:
                print('Warning, no symmetry relation found!')
            if i_sym > i_sym_max:
                i_sym_max = i_sym
        i_sym_max += 1

        return i_sym_max

    def build_symmetry_map(self):
        """Build the dictionary mapsym (i -> j) that associate one edge j per group of symmetry i."""

        mapsym = {}

        i = 0
        while True:
            j = 0
            has_i_sym = False
            for u, v in self.edges():
                i_sym = self.edge_attribute((u, v), 'sym_key')
                if i_sym == i:
                    mapsym[i] = j
                    has_i_sym = True
                    break
                j += 1
            if has_i_sym is False:
                break
            i += 1

        return mapsym

    # --------------------------------------------------------------- #
    # -----------------------TNA-CONECTION--------------------------- #
    # --------------------------------------------------------------- #

    def build_dual(self):
        """
        Build the dual of the form diagram.

        Returns
        -------
        dual
            The middle dual mesh.

        Notes
        -----
        Construction of the dual diagram is based on the faces around the inner, free vertices of the form diagram.
        This means not only the vertices on the boundary are ignored, but also the vertices that are anchored.

        """

        self.modify_diagram_boundary()
        dual = self.__class__.dual(self, Mesh)

        return dual

    def build_complete_dual(self):
        """
        Build the dual of the form diagram considering half-faces in the edges on the perimeter.

        Returns
        -------
        dual
            The middle dual mesh.

        Notes
        -----
        Construction of the dual diagram is based on the faces around all edges.
        The edges in the boundary recieve special tratment and their midpoint is.

        """
        dual = Mesh()
        fkey_centroid = {fkey: self.face_centroid(fkey) for fkey in self.faces()}
        midpt_boundary = {(u, v): self.edge_midpoint(u, v) for u, v in self.edges()}
        fixed = self.fixed()
        boundary = list(self.vertices_on_boundary())
        vertices = {}
        faces = {}
        i = 0  # face keys
        j = 0  # vertex keys
        for key in self.vertices():
            facekeys = []
            fkeys = self.vertex_faces(key, ordered=True)
            for fkey in fkeys:
                pt = fkey_centroid[fkey]
                if pt not in vertices.values():
                    vertices[j] = pt
                    facekeys.append(j)
                    j += 1
                else:
                    j_back = list(vertices.keys())[list(vertices.values()).index(pt)]
                    facekeys.append(j_back)
            if key in boundary:
                nbrs = self.vertex_neighbors(key, ordered=True)
                nbrs.reverse()
                for nbr in nbrs:
                    if nbr in boundary:
                        if (key, nbr) in midpt_boundary.keys():
                            pt = midpt_boundary[(key, nbr)]
                        else:
                            pt = midpt_boundary[(nbr, key)]

                        if pt not in vertices.values():
                            vertices[j] = pt
                            facekeys.append(j)
                            j += 1
                        else:
                            j_back = list(vertices.keys())[list(vertices.values()).index(pt)]
                            facekeys.append(j_back)
            if key in fixed:
                pt = self.vertex_coordinates(key)
                vertices[j] = pt
                facekeys = facekeys[:-1] + [j] + [facekeys[-1]]  # add the support as last element
                j += 1

            faces[i] = facekeys
            i += 1

        for key, (x, y, z) in vertices.items():
            dual.add_vertex(key, x=x, y=y, z=z)
        for fkey, vertices in faces.items():
            dual.add_face(vertices, fkey=fkey)

        return dual

    def modify_diagram_boundary(self):
        """
        Modify the diagram boundary adding the outer perimeter around the fixed vertices with edges having ``_is_edge`` False.

        Parameters
        ----------
        delete_face : bool, optional
            Option to delete the outside face of the diagram.
            The default value is ``False``.

        Returns
        -------
        None
            The ``FormDiagram`` is modified in place.

        """

        corners = list(self.vertices_where({'is_fixed': True}))
        self.vertices_attributes(('is_anchor', 'is_fixed'), (True, True), keys=corners)
        self.update_boundaries()

        return

    def restore_diagram_boundary(self, tol=1e-3):
        """
        Remove faces of the diagram with area zero.


        Returns
        -------
        None
            The ``FormDiagram`` is modified in place.

        """

        faces_delete = []
        for face in self.faces():
            A = self.face_area(face)
            if A < tol:
                faces_delete.append(face)

        for face in faces_delete:
            self.delete_face(face)

        return





from numpy import zeros
from numpy import ones
from numpy import array
from numpy import linspace

from compas_tno.shapes import MeshDos
from compas_tno.shapes import rectangular_topology

import math


def pavillion_vault_highfields_proxy(xy_span, thk=0.5, discretisation=10, t=0.0):
    """Get Pavillionvault from proxy.

    Parameters
    ----------
    xy_span : [list, list], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    thk : float, optional
        The thickness of the vault, by default 0.5
    tol : float, optional
        Tolerance, by default 10e-6
    t : float, optional
        Parameter for lower bound in nodes in the boundary, by default 0.0
    discretisation : list|int, optional
        Level of discretisation of the shape, by default [100, 100]

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the shape
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the shape
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the shape
    """

    intrados, extrados, middle = pavillion_vault_highfields(xy_span=xy_span, thk=thk, discretisation=discretisation, t=t)
    return intrados.to_data(), extrados.to_data(), middle.to_data()


def pavillion_vault_highfields(xy_span=[[0.0, 10.0], [0.0, 10.0]], thk=None, tol=10e-6, t=0.0, discretisation=[100, 100], spr_angle=0.0, expanded=False):
    """Set Pavillion vault heights

    Parameters
    ----------
    xy_span : [list, list], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    thk : float, optional
        The thickness of the vault, by default 0.5
    tol : float, optional
        Tolerance, by default 10e-6
    t : float, optional
        Parameter for lower bound in nodes in the boundary, by default 0.0
    discretisation : list|int, optional
        Level of discretisation of the shape, by default [100, 100]
    spr_angle : float, optional
        Springing angle, by default 0.0
    expanded : bool, optional
        If the extrados should extend beyond the floor plan, by default False

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the shape
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the shape
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the shape
    """

    # TODO: Update this function to work on rectangular Pavillion-Vaults

    if isinstance(discretisation, int):
        discretisation = [discretisation, discretisation]

    y1 = xy_span[1][1]
    y0 = xy_span[1][0]
    x1 = xy_span[0][1]
    x0 = xy_span[0][0]

    density_x = discretisation[0]
    density_y = discretisation[1]
    x = linspace(x0, x1, num=density_x+1, endpoint=True)  # arange(x0, x1 + dx/density_x, dx/density_x)
    y = linspace(y0, y1, num=density_y+1, endpoint=True)  # arange(y0, y1 + dy/density_y, dy/density_y)

    xi, yi, faces_i = rectangular_topology(x, y)

    zt = pavillionvault_middle_update(xi, yi, xy_span=xy_span, spr_angle=spr_angle, tol=1e-6)
    xyzt = array([xi, yi, zt.flatten()]).transpose()
    middle = MeshDos.from_vertices_and_faces(xyzt, faces_i)

    zub, zlb = pavillionvault_ub_lb_update(xi, yi, thk, t, xy_span=xy_span, spr_angle=spr_angle, tol=1e-6)

    xyzub = array([xi, yi, zub.flatten()]).transpose()
    xyzlb = array([xi, yi, zlb.flatten()]).transpose()

    extrados = MeshDos.from_vertices_and_faces(xyzub, faces_i)
    intrados = MeshDos.from_vertices_and_faces(xyzlb, faces_i)

    if expanded:
        x = linspace(x0 - thk/2 / math.cos(math.radians(spr_angle)), x1 + thk/2 / math.cos(math.radians(spr_angle)),
                     num=density_x+1, endpoint=True)  # arange(x0, x1 + dx/density_x, dx/density_x)
        y = linspace(y0 - thk/2 / math.cos(math.radians(spr_angle)), y1 + thk/2 / math.cos(math.radians(spr_angle)),
                     num=density_y+1, endpoint=True)  # arange(y0, y1 + dy/density_y, dy/density_y)
        xi, yi, faces_i = rectangular_topology(x, y)
        zub, zlb = pavillionvault_ub_lb_update(xi, yi, thk, t, xy_span=xy_span, spr_angle=spr_angle, tol=1e-6)
        xyzub = array([xi, yi, zub.flatten()]).transpose()
        extrados = MeshDos.from_vertices_and_faces(xyzub, faces_i)

    return intrados, extrados, middle


def pavillionvault_middle_update(x, y, xy_span=[[0.0, 10.0], [0.0, 10.0]], spr_angle=0.0, tol=1e-6):
    """Update middle of a pavillion vault based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the vault
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    spr_angle : float, optional
        Springing angle, by default 0.0
    tol : float, optional
        Tolerance, by default 10e-6

    Returns
    -------
    z : array
        Values of the middle surface in the points
    """

    x0, x1 = xy_span[0]
    y0, y1 = xy_span[1]

    if spr_angle == 0.0:
        z_ = 0.0
    else:
        alpha = 1/math.cos(math.radians(spr_angle))
        z_ = (x1 - x0)/2 * math.tan(math.radians(spr_angle))
        L = x1 * alpha
        Ldiff = L - x1
        x0, x1 = -Ldiff/2, x1 + Ldiff/2
        y0, y1 = -Ldiff/2, y1 + Ldiff/2

    rx = (x1 - x0)/2
    ry = (y1 - y0)/2

    z = zeros((len(x), 1))

    for i in range(len(x)):
        xi, yi = x[i], y[i]
        if (yi - y0) <= y1/x1 * (xi - x0) + tol and (yi - y0) <= (y1 - y0) - (xi - x0) + tol:  # Q1
            z[i] = math.sqrt((ry)**2 - ((yi - y0)-ry)**2) - z_
        elif (yi - y0) >= y1/x1 * (xi - x0) - tol and (yi - y0) >= (y1 - y0) - (xi - x0) - tol:  # Q3
            z[i] = math.sqrt((ry)**2 - ((yi - y0)-ry)**2) - z_
        elif (yi - y0) <= y1/x1 * (xi - x0) + tol and (yi - y0) >= (y1 - y0) - (xi - x0) - tol:  # Q2
            z[i] = math.sqrt((rx)**2 - ((xi - x0)-rx)**2) - z_
        elif (yi - y0) >= y1/x1 * (xi - x0) - tol and (yi - y0) <= (y1 - y0) - (xi - x0) + tol:  # Q4
            z[i] = math.sqrt((rx)**2 - ((xi - x0)-rx)**2) - z_
        else:
            print('Error Q. (x,y) = ({0},{1})'.format(xi, yi))

    return z


def pavillionvault_ub_lb_update(x, y, thk, t, xy_span=[[0.0, 10.0], [0.0, 10.0]], spr_angle=0.0, tol=1e-6):
    """Update upper and lower bounds of a pavillionvault based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the vault
    t : float
        Parameter for lower bound in nodes in the boundary
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    tol : float, optional
        Tolerance, by default 10e-6

    Returns
    -------
    ub : array
        Values of the upper bound in the points
    lb : array
        Values of the lower bound in the points
    """

    x0, x1 = xy_span[0]
    y0, y1 = xy_span[1]

    if spr_angle == 0.0:
        z_ = 0.0
    else:
        alpha = 1/math.cos(math.radians(spr_angle))
        z_ = (x1 - x0)/2 * math.tan(math.radians(spr_angle))
        L = x1 * alpha
        Ldiff = L - x1
        x0, x1 = -Ldiff/2, x1 + Ldiff/2
        y0, y1 = -Ldiff/2, y1 + Ldiff/2

    y1_ub = y1 + thk/2
    y0_ub = y0 - thk/2
    x1_ub = x1 + thk/2
    x0_ub = x0 - thk/2

    y1_lb = y1 - thk/2
    y0_lb = y0 + thk/2
    x1_lb = x1 - thk/2
    x0_lb = x0 + thk/2

    rx_ub = (x1_ub - x0_ub)/2
    ry_ub = (y1_ub - y0_ub)/2
    rx_lb = (x1_lb - x0_lb)/2
    ry_lb = (y1_lb - y0_lb)/2

    ub = ones((len(x), 1))
    lb = ones((len(x), 1)) * - t

    for i in range(len(x)):
        xi, yi = x[i], y[i]
        intrados_null = False
        if yi > y1_lb or xi > x1_lb or xi < x0_lb or yi < y0_lb:
            intrados_null = True
        if (yi - y0) <= y1/x1 * (xi - x0) + tol and (yi - y0) <= (y1 - y0) - (xi - x0) + tol:  # Q1
            ub[i] = math.sqrt((ry_ub)**2 - ((yi - y0_ub)-ry_ub)**2) - z_
            if not intrados_null:
                lb[i] = math.sqrt((ry_lb)**2 - ((yi - y0_lb)-ry_lb)**2) - z_
        elif (yi - y0) >= y1/x1 * (xi - x0) - tol and (yi - y0) >= (y1 - y0) - (xi - x0) - tol:  # Q3
            ub[i] = math.sqrt((ry_ub)**2 - ((yi - y0_ub)-ry_ub)**2) - z_
            if not intrados_null:
                lb[i] = math.sqrt((ry_lb)**2 - ((yi - y0_lb)-ry_lb)**2) - z_
        elif (yi - y0) <= y1/x1 * (xi - x0) + tol and (yi - y0) >= (y1 - y0) - (xi - x0) - tol:  # Q2
            ub[i] = math.sqrt((rx_ub)**2 - ((xi - x0_ub)-rx_ub)**2) - z_
            if not intrados_null:
                lb[i] = math.sqrt((rx_lb)**2 - ((xi - x0_lb)-rx_lb)**2) - z_
        elif (yi - y0) >= y1/x1 * (xi - x0) - tol and (yi - y0) <= (y1 - y0) - (xi - x0) + tol:  # Q4
            ub[i] = math.sqrt((rx_ub)**2 - ((xi - x0_ub)-rx_ub)**2) - z_
            if not intrados_null:
                lb[i] = math.sqrt((rx_lb)**2 - ((xi - x0_lb)-rx_lb)**2) - z_
        else:
            print('Error Q. (x,y) = ({0},{1})'.format(xi, yi))

    return ub, lb


def pavillionvault_dub_dlb(x, y, thk, t, xy_span=[[0.0, 10.0], [0.0, 10.0]], tol=1e-6):
    """Computes the sensitivities of upper and lower bounds in the x, y coordinates and thickness specified.

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the vault
    t : float
        Parameter for lower bound in nodes in the boundary
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    tol : float, optional
        Tolerance, by default 10e-6

    Returns
    -------
    dub : array
        Values of the sensitivities for the upper bound in the points
    dlb : array
        Values of the sensitivities for the lower bound in the points
    """

    x0, x1 = xy_span[0]
    y0, y1 = xy_span[1]

    y1_ub = y1 + thk/2
    y0_ub = y0 - thk/2
    x1_ub = x1 + thk/2
    x0_ub = x0 - thk/2

    y1_lb = y1 - thk/2
    y0_lb = y0 + thk/2
    x1_lb = x1 - thk/2
    x0_lb = x0 + thk/2

    rx_ub = (x1_ub - x0_ub)/2
    ry_ub = (y1_ub - y0_ub)/2
    rx_lb = (x1_lb - x0_lb)/2
    ry_lb = (y1_lb - y0_lb)/2

    ub = ones((len(x), 1))
    lb = ones((len(x), 1)) * - t
    dub = zeros((len(x), 1))
    dlb = zeros((len(x), 1))

    for i in range(len(x)):
        xi, yi = x[i], y[i]
        intrados_null = False
        if yi > y1_lb or xi > x1_lb or xi < x0_lb or yi < y0_lb:
            intrados_null = True
        if (yi - y0) <= y1/x1 * (xi - x0) + tol and (yi - y0) <= (y1 - y0) - (xi - x0) + tol:  # Q1
            ub[i] = math.sqrt((ry_ub)**2 - ((yi - y0_ub)-ry_ub)**2)
            dub[i] = 1/2 * ry_ub/ub[i]
            if not intrados_null:
                lb[i] = math.sqrt((ry_lb)**2 - ((yi - y0_lb)-ry_lb)**2)
                dlb[i] = - 1/2 * ry_lb/lb[i]
        elif (yi - y0) >= y1/x1 * (xi - x0) - tol and (yi - y0) >= (y1 - y0) - (xi - x0) - tol:  # Q3
            ub[i] = math.sqrt((ry_ub)**2 - ((yi - y0_ub)-ry_ub)**2)
            dub[i] = 1/2 * ry_ub/ub[i]
            if not intrados_null:
                lb[i] = math.sqrt((ry_lb)**2 - ((yi - y0_lb)-ry_lb)**2)
                dlb[i] = - 1/2 * ry_lb/lb[i]
        elif (yi - y0) <= y1/x1 * (xi - x0) + tol and (yi - y0) >= (y1 - y0) - (xi - x0) - tol:  # Q2
            ub[i] = math.sqrt((rx_ub)**2 - ((xi - x0_ub)-rx_ub)**2)
            dub[i] = 1/2 * rx_ub/ub[i]
            if not intrados_null:
                lb[i] = math.sqrt((rx_lb)**2 - ((xi - x0_lb)-rx_lb)**2)
                dlb[i] = - 1/2 * rx_lb/lb[i]
        elif (yi - y0) >= y1/x1 * (xi - x0) - tol and (yi - y0) <= (y1 - y0) - (xi - x0) + tol:  # Q4
            ub[i] = math.sqrt((rx_ub)**2 - ((xi - x0_ub)-rx_ub)**2)
            dub[i] = 1/2 * rx_ub/ub[i]
            if not intrados_null:
                lb[i] = math.sqrt((rx_lb)**2 - ((xi - x0_lb)-rx_lb)**2)
                dlb[i] = - 1/2 * rx_lb/lb[i]
        else:
            print('Error Q. (x,y) = ({0},{1})'.format(xi, yi))

    return dub, dlb  # ub, lb


def pavillionvault_b_update(x, y, thk, fixed, xy_span=[[0.0, 10.0], [0.0, 10.0]]):
    """Computes the ``b`` of parameter x, y coordinates and thickness specified.

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the vault
    fixed : list
        The list with indexes of the fixed vertices
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]

    Returns
    -------
    dub : array
        Values of the sensitivities for the upper bound in the points
    dlb : array
        Values of the sensitivities for the lower bound in the points
    """

    x0, x1 = xy_span[0]
    y0, y1 = xy_span[1]
    b = zeros((len(fixed), 2))

    for i in range(len(fixed)):
        index = fixed[i]
        xi, yi = x[index], y[index]
        if xi == x0:
            b[[i], :] += [-thk/2, 0]
        elif xi == x1:
            b[i, :] += [+thk/2, 0]
        if yi == y0:
            b[i, :] += [0, -thk/2]
        elif yi == y1:
            b[i, :] += [0, +thk/2]

    return abs(b)


def pavillionvault_db(x, y, thk, fixed, xy_span=[[0.0, 10.0], [0.0, 10.0]]):
    """Computes the sensitivities of the ``b`` parameter in the x, y coordinates and thickness specified.

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the vault
    fixed : list
        The list with indexes of the fixed vertices
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]

    Returns
    -------
    db : array
        Values of the sensitivities of the ``b`` parameter in the points
    """

    x0, x1 = xy_span[0]
    y0, y1 = xy_span[1]
    db = zeros((len(fixed), 2))

    for i in range(len(fixed)):
        index = fixed[i]
        xi, yi = x[index], y[index]
        if xi == x0:
            db[i, :] += [-1/2, 0]
        elif xi == x1:
            db[i, :] += [+1/2, 0]
        if yi == y0:
            db[i, :] += [0, -1/2]
        elif yi == y1:
            db[i, :] += [0, +1/2]

    return abs(db)



from compas_tno.diagrams import ForceDiagram

from compas_tna.equilibrium import horizontal_nodal

from compas_tno.algorithms.equilibrium import vertical_equilibrium_fdm
from compas_tno.plotters import TNOPlotter

from compas.numerical import connectivity_matrix
from compas.numerical import spsolve_with_known

from numpy import array
from numpy import float64

from scipy.sparse import diags


def form_update_with_parallelisation(form, zmax=5.0, plot=False, printout=False, alpha=100.0, kmax=500, callback=None):
    """Parallelise a TNO form diagram using as starting point the centroidal dual as ForceDiagram.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram to update.
    zmax : float, optional
        The maximum height of a node in the thrust network.
        The default value is ``5.0``.
    plot : bool, optional
        If plots with form and force should display on the screen.
        The default value is ``False``.
    plot : bool, optional
        If prints should appear in the screen with key information.
        The default value is ``False``.
    alpha : float, optional
        Weighting factor for computation of the target vectors (the default is
        100.0, which implies that the target vectors are the edges of the form diagram).
        If 0.0, the target vectors are the edges of the force diagram.
    kmax : int, optional
       Maximum number of iterations.
       Default is ``500``.
    callback : callable, optional
        A callback function to be called at every iteration of the parallelisation algorithm.
        The callback should take the current iterand, the coordinates of the form diagram,
        and the coordinates of the force diagram as input parameters.
        Default is ``None``.

    Nores
    ---------
    See `compas tna <https://blockresearchgroup.github.io/compas_tna/>`_  package.

    Returns
    -------
    force: ForceDiagram
        The force diagram after the process.
    """
    # mark supports as 'is_anchor'
    corners = list(form.vertices_where({'is_fixed': True}))
    form.vertices_attribute('is_anchor', True, keys=corners)

    # update bounds adding a new face to the open edges along boundaries
    # does not apply for continuous boundary cases, for such, the edges
    # on the boundary must be previously set with '_is_edge': False
    leaves = False
    for u, v in form.edges_on_boundary():
        if form.edge_attribute((u, v), '_is_edge') is False:
            leaves = True
            break
    if leaves is False:
        form.update_boundaries()
    force = ForceDiagram.from_formdiagram(form)

    # ad = '/Users/mricardo/compas_dev/compas_tno/data/CISM/forces/force-0.json'
    # force.to_json(ad)

    # for force_edge in force.edges():
    #     form_edge = force.dual_edge(force_edge)
    #     q = form.edge_attribute(form_edge, 'q')
    #     p1, p2 = form.edge_coordinates(*form_edge)
    #     lh = distance_point_point_xy(p1, p2)
    #     lmin = lmax = abs(q * lh)/10
    #     force.edge_attribute(force_edge, 'lmax', lmax)
    #     # if form.edge_attribute(form_edge, 'is_ind'):
    #     #     force.edge_attribute(force_edge, 'lmin', lmin)

    # def callback(k, xyz_force, edges):
    #     print('Iteration:', k)
    #     for i, key in enumerate(force.vertices()):
    #         force.vertex_attribute(key, 'x', xyz_force[i][0])
    #         force.vertex_attribute(key, 'y', xyz_force[i][1])

    #     # ad = '/Users/mricardo/compas_dev/compas_tno/data/CISM/forces/force-' + str(k + 1) + '.json'
    #     # force.to_json(ad)

    #     plotter = TNOPlotter(force=force)
    #     plotter.draw_force()
    #     plotter.show()

    # plot of diagrams @ initial state
    if plot:
        plotter = TNOPlotter(form=form, force=force, figsize=(16, 8))
        plotter.draw_mesh()
        plotter.draw_force()
        plotter.show()

    # horizontal equilibrium
    horizontal_nodal(form, force, alpha=alpha, kmax=kmax, callback=callback)

    if printout:
        dev = form.edges_attribute('_a')
        print('max/min deviations:', max(dev), min(dev))

    # ****** change compression to negative (TNO convention)
    q = [form.edge_attribute((u, v), 'q') for u, v in form.edges_where({'_is_edge': True})]
    for index, edge in enumerate(form.edges_where({'_is_edge': True})):
        form.edge_attribute(edge, 'q', -1 * q[index])

    # update vertical equilibrium (only z coordinates change)
    vertical_equilibrium_fdm(form, zmax=zmax)

    # plot of diagrams @ final state
    if plot:
        plotter = TNOPlotter(form=form, force=force, figsize=(16, 8))
        plotter.draw_mesh()
        plotter.draw_force()
        plotter.show()

    return force


def reciprocal_from_form(form, plot=False, restore_form_topology=True):
    """Find a reciprocal form diagram from the stored force densities in thisdiagram
    This is an algebraic solution to the duality. It should work as long as the force densities
    in the form diagram are equilibrated.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram to update.
    plot : bool, optional
        If plots with form and force should display on the screen.
        The default value is ``False``.
    restore_form_topology : bool, optional
        If form topology should be restored.
        The default value is ``True``.

    Notes
    ---------
    See `compas ags <https://blockresearchgroup.github.io/compas_ags/>`_  package.

    Returns
    -------
    force : :class:`ForceDiagram`
        The reciprocal force diagram.

    """
    # mark supports as 'is_anchor'
    corners = list(form.vertices_where({'is_fixed': True}))
    form.vertices_attribute('is_anchor', True, keys=corners)

    # update bounds adding a new face to the open edges along boundaries
    # does not apply for continuous boundary cases, for such, the edges
    # on the boundary must be previously set with '_is_edge': False
    leaves = False
    for u, v in form.edges_on_boundary():
        if form.edge_attribute((u, v), '_is_edge') is False:
            leaves = True
            break
    if leaves is False:
        init_faces = list(form.faces())
        form.update_boundaries()

    # print('edges', form.number_of_edges())
    # print('vertices', form.number_of_vertices())
    # print('faces', form.number_of_faces())

    force = ForceDiagram.from_formdiagram(form)

    # print('force edges', force.number_of_edges())
    # print('force vertices', force.number_of_vertices())
    # print('force faces', force.number_of_faces())

    if plot:
        plotter = TNOPlotter(form=form, force=force, figsize=(16, 8))
        plotter.draw_mesh()
        plotter.draw_force()
        plotter.show()

    force_update_from_form(force, form)

    # this modifies "back" the form diagram in case open edges in the boundary are present
    # It will delete the added faces above.
    if restore_form_topology:
        if leaves is False:
            end_faces = list(form.faces())
            new_faces = list(set(end_faces) - set(init_faces))
            for fkey in new_faces:
                form.delete_face(fkey)

    if plot:
        print('Plot of Reciprocal')
        force.plot()

    return force


def force_update_from_form(force, form):
    """Update the force diagram after modifying the (force densities of) the form diagram.
    This is an algebraic solution to the duality. It works as long as the force densities
    in the form diagram are equilibrated.

    Parameters
    ----------
    force : :class:`~compas_tno.diagrams.ForceDiagram`
        The force diagram on which the update is based.
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram to update.

    Notes
    ---------
    See `compas tna <https://blockresearchgroup.github.io/compas_tna/>`_ package.

    Returns
    -------
    None
        The form and force diagram are updated in-place.
    """
    # --------------------------------------------------------------------------
    # form diagram
    # --------------------------------------------------------------------------
    vertex_index = form.vertex_index()

    xy = array(form.xy(), dtype=float64)
    edges = [[vertex_index[u], vertex_index[v]] for u, v in form.edges_where({'_is_edge': True})]
    C = connectivity_matrix(edges, 'csr')
    Q = diags([form.q()], [0])
    uv = C.dot(xy)
    # --------------------------------------------------------------------------
    # force diagram
    # --------------------------------------------------------------------------
    _vertex_index = force.vertex_index()

    _known = [_vertex_index[force.anchor()]]
    _xy = array(force.xy(), dtype=float64)
    _edges = force.ordered_edges(form)
    _edges[:] = [(_vertex_index[u], _vertex_index[v]) for u, v in _edges]
    _C = connectivity_matrix(_edges, 'csr')
    _Ct = _C.transpose()
    # --------------------------------------------------------------------------
    # compute reciprocal for given q
    # --------------------------------------------------------------------------
    _xy = spsolve_with_known(_Ct.dot(_C), _Ct.dot(Q).dot(uv), _xy, _known)
    # --------------------------------------------------------------------------
    # rotate force diagram to make it parallel to the form diagram
    # use CCW direction (opposite of cycle direction)
    # --------------------------------------------------------------------------
    _x, _y = zip(*_xy)
    _xy[:] = [list(item) for item in zip([-_ for _ in _y], _x)]
    # --------------------------------------------------------------------------
    # rotate the force diagram 90 degrees in CW direction
    # this way the relation between the two diagrams is easier to read
    # --------------------------------------------------------------------------
    # _x, _y = zip(*_xy)
    # _xy[:] = [list(item) for item in zip(_y, [-_ for _ in _x])]
    # --------------------------------------------------------------------------
    # update force diagram
    # --------------------------------------------------------------------------
    for vertex, attr in force.vertices(True):
        index = _vertex_index[vertex]
        attr['x'] = _xy[index, 0]
        attr['y'] = _xy[index, 1]
        attr['z'] = 0.0
